// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package main

import (
	"encoding/json"
	"fmt"
	"io"
	"log" // Note: *not* using internal/log here because we don't want its formatting
	"net"
	"os"
	"sort"
	"strconv"
	"strings"
	"time"

	"github.com/integrii/flaggy"
	"github.com/olekukonko/tablewriter"
	"sr.ht/~craftyguy/superd/internal/color"
	"sr.ht/~craftyguy/superd/pkg/api"
)

// set at build time
var Version string

var logger *log.Logger

func fatalError(msg string) {
	logger.Fatal("error received from superd:\n" + msg)
}

func main() {
	// disable timestamps, etc in logger output
	logger = log.Default()
	logger.SetFlags(0)

	// this depends on XDG_RUNTIME_DIR, so make sure it is set and fail fast if
	// not
	if v, ok := os.LookupEnv("XDG_RUNTIME_DIR"); v == "" || !ok {
		logger.Fatal("$XDG_RUNTIME_DIR is not set in the environment, this is required.")
	}

	flaggy.SetName("superctl")
	flaggy.SetDescription("Control app for the superd supervisor")
	flaggy.SetVersion(Version)

	// Subcommands
	startCmd := flaggy.NewSubcommand("start")
	startCmd.Description = "Start the given service"
	flaggy.AttachSubcommand(startCmd, 1)

	stopCmd := flaggy.NewSubcommand("stop")
	stopCmd.Description = "Stop the given service"
	flaggy.AttachSubcommand(stopCmd, 1)

	restartCmd := flaggy.NewSubcommand("restart")
	restartCmd.Description = "Restart the given service"
	flaggy.AttachSubcommand(restartCmd, 1)

	enableCmd := flaggy.NewSubcommand("enable")
	enableCmd.Description = "Enable the given service"
	flaggy.AttachSubcommand(enableCmd, 1)

	disableCmd := flaggy.NewSubcommand("disable")
	disableCmd.Description = "Disable the given service"
	flaggy.AttachSubcommand(disableCmd, 1)

	statusCmd := flaggy.NewSubcommand("status")
	statusCmd.Description = "Get status for the give service(s), or all services if not specified"
	flaggy.AttachSubcommand(statusCmd, 1)

	setEnvCmd := flaggy.NewSubcommand("set-env")
	setEnvCmd.Description = "Set one or more environment variables for running services, either to the value specified or the value from the current environment if it is set"
	flaggy.AttachSubcommand(setEnvCmd, 1)

	reloadCmd := flaggy.NewSubcommand("reload")
	reloadCmd.Description = "Reload service config from disk, adding any new config found"
	flaggy.AttachSubcommand(reloadCmd, 1)

	// Subcommand flags
	var now bool
	enableCmd.Bool(&now, "n", "now", "Start service after enabling it")
	disableCmd.Bool(&now, "n", "now", "Stop service after disabling it")

	// Subcommand positional args
	var service string
	startCmd.AddPositionalValue(&service, "service", 1, true, "Service to start")
	stopCmd.AddPositionalValue(&service, "service", 1, true, "Service to stop")
	restartCmd.AddPositionalValue(&service, "service", 1, true, "Service to restart")
	enableCmd.AddPositionalValue(&service, "service", 1, true, "Service to enable")
	disableCmd.AddPositionalValue(&service, "service", 1, true, "Service to disable")
	statusCmd.AddPositionalValue(&service, "service", 1, false, "Service to get status for, or all if not specified")

	// allows for parsing trailing args:
	flaggy.ShowHelpOnUnexpectedDisable()

	flaggy.Parse()

	socketPath := os.ExpandEnv("$XDG_RUNTIME_DIR/superd.sock")

	if startCmd.Used {

		startService(service, socketPath)

	} else if stopCmd.Used {

		stopService(service, socketPath)

	} else if restartCmd.Used {
		var resp api.NilResponse

		resBuf, err := sendCmd(socketPath, api.CmdRestart, []string{service})
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {
			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

	} else if statusCmd.Used {
		var resp api.ServiceStatusResponse

		// if service is an empty string then the user wanted status for all
		// services, so make sure to send an empty list (and not a list with an
		// empty string...) to superd
		var arg []string
		if service != "" {
			arg = append(arg, service)
		}

		resBuf, err := sendCmd(socketPath, api.CmdStatus, arg)
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {

			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

		if len(resp.Result) == 0 {
			logger.Fatal("no statuses were returned by superd")
		} else if service == "" || len(resp.Result) > 1 {
			// print less detailed status for many services
			if err := printStatusMulti(resp.Result); err != nil {
				logger.Fatal(err)
			}
		} else {
			// print detailed output for 1 service
			if err := printStatusSingle(service, resp.Result[service]); err != nil {
				logger.Fatal(err)
			}
		}

	} else if enableCmd.Used {
		var resp api.StringResponse

		args := []string{service}
		resBuf, err := sendCmd(socketPath, api.CmdEnable, args)
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {
			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

		logger.Println(resp.Result)

		if now {
			startService(service, socketPath)
		}

	} else if disableCmd.Used {
		var resp api.StringResponse

		args := []string{service}
		resBuf, err := sendCmd(socketPath, api.CmdDisable, args)
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {
			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

		logger.Println(resp.Result)

		if now {
			stopService(service, socketPath)
		}

	} else if setEnvCmd.Used {
		var resp api.NilResponse

		args := []string{}

		// set values for any env vars where the value was not passed on the
		// cmd line
		for _, v := range flaggy.TrailingArguments {
			if !strings.ContainsRune(v, '=') {
				// if it's unset in the current env, just allow it to be set to
				// an empty string. i.e. the user may want the var to still be
				// present in the environment.
				v = fmt.Sprintf("%s=%s", v, os.Getenv(v))
			}
			args = append(args, v)
		}

		resBuf, err := sendCmd(socketPath, api.CmdSetEnv, args)
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {
			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

	} else if reloadCmd.Used {
		var resp api.NilResponse

		resBuf, err := sendCmd(socketPath, api.CmdReload, []string{})
		if err != nil {
			logger.Fatal(err)
		}

		if err := json.Unmarshal(resBuf, &resp); err != nil {
			logger.Fatal(err)
		}

		if resp.Error {
			fatalError(resp.ErrorMsg)
		}

	} else {
		flaggy.ShowHelpAndExit("a subcommand is required")
	}
}

func startService(service string, socketPath string) {

	var resp api.NilResponse

	resBuf, err := sendCmd(socketPath, api.CmdStart, []string{service})
	if err != nil {
		logger.Fatal(err)
	}

	if err := json.Unmarshal(resBuf, &resp); err != nil {
		logger.Fatal(err)
	}

	if resp.Error {
		fatalError(resp.ErrorMsg)
	}
}

func stopService(service string, socketPath string) {

	resBuf, err := sendCmd(socketPath, api.CmdStop, []string{service})
	if err != nil {
		logger.Fatal(err)
	}

	var resp api.NilResponse
	if err := json.Unmarshal(resBuf, &resp); err != nil {
		logger.Fatal(err)
	}

	if resp.Error {
		fatalError(resp.ErrorMsg)
	}
}

func sendCmd(socketPath string, cmd api.ReqCommand, args []string) (resp []byte, err error) {
	sock, err := net.Dial("unix", socketPath)
	if err != nil {
		err = fmt.Errorf("unable to open socket, is superd running?\n%w", err)
		return
	}
	defer sock.Close()

	req := api.Request{
		Cmd:    cmd,
		Params: args,
	}

	reqBuf, err := req.Json()
	if err != nil {
		return
	}

	if err = sock.SetReadDeadline(time.Now().Add(2 * time.Second)); err != nil {
		return
	}

	if _, err = sock.Write(reqBuf); err != nil {
		return
	}

	if err = sock.SetDeadline(time.Now().Add(15 * time.Second)); err != nil {
		return
	}

	resp, err = io.ReadAll(sock)

	return
}

func colorState(state api.ServiceState) string {
	switch state {
	case api.ServiceFailed:
		return color.Red(state)
	case api.ServiceStarted:
		return color.Green(state)
	case api.ServiceStopped:
		return color.Yellow(state)
	default:
		return string(state)
	}
}

func printStatusMulti(services map[string]api.ServiceStatus) error {
	// sort services in response
	var keys []string
	for k := range services {
		keys = append(keys, k)
	}
	sort.Strings(keys)

	rows := [][]string{}

	for _, k := range keys {
		var (
			state  string
			pid                      = "-"
			auto                     = "-"
			status api.ServiceStatus = services[k]
		)

		state = colorState(status.State)
		if status.State == api.ServiceStarted {
			if status.Type != "oneshot" {
				// info that doesn't make sense when type is oneshot
				pid = strconv.Itoa((status.Pid))
			}
		}

		if status.Autostart {
			auto = "✓"
		}

		rows = append(rows, []string{k, state, pid, auto})
	}

	table := newTable()
	table.SetHeader([]string{"", "Status", "PID", "Autostart"})
	table.AppendBulk(rows)
	table.Render()
	return nil
}

func printStatusSingle(name string, status api.ServiceStatus) error {

	var (
		state  string
		auto   = "-"
		pid    = "-"
		uptime = "-"
	)

	state = colorState(status.State)
	if status.State == api.ServiceStarted {
		// skip info that doesn't make sense when type is oneshot
		if status.Type != "oneshot" {
			pid = strconv.Itoa(status.Pid)
			uptime = time.Duration(float64(status.Uptime) * float64(time.Second)).String()
		}
	}

	if status.Autostart {
		auto = "✓"
	}

	table := newTable()
	table.AppendBulk([][]string{
		{"", "Status:", state},
		{"", "Type:", status.Type},
		{"", "PID:", pid},
		{"", "Uptime:", uptime},
		{"", "Autostart:", auto},
		{"", "Loaded:", status.ConfigPath},
		{"", "Description:", status.Description},
	})
	fmt.Println(name)
	table.Render()

	return nil
}

// helper that returns a table configured to display as little
// decoration/border/padding as possible
func newTable() (table *tablewriter.Table) {
	table = tablewriter.NewWriter(os.Stdout)
	table.SetAutoWrapText(false)
	table.SetRowSeparator("")
	table.SetColumnSeparator("")
	table.SetCenterSeparator("")
	table.SetHeaderLine(false)
	table.SetHeaderAlignment(tablewriter.ALIGN_LEFT)
	table.SetAlignment(tablewriter.ALIGN_LEFT)
	table.SetAutoFormatHeaders(false) // disables all caps for header text
	table.SetBorder(false)
	table.SetNoWhiteSpace(true)
	table.SetTablePadding("   ") // separator between columns
	return table
}
