// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package main

import (
	"fmt"
	"io"
	"os"
	"path/filepath"
	"testing"
	"time"

	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/internal/manager"
	superdTesting "sr.ht/~craftyguy/superd/internal/testing"
	"sr.ht/~craftyguy/superd/pkg/api"
)

func copyFile(src string, dest string) (err error) {
	in, err := os.Open(src)
	if err != nil {
		return
	}
	defer in.Close()

	out, err := os.Create(dest)
	if err != nil {
		return
	}
	defer out.Close()

	_, err = io.Copy(out, in)

	return
}

// manager is instantiated once and re-used by all tests, since it has a
// startup cost that can really slow down testing if a new manager is started
// each time.
var mgr *manager.Manager

func setup() (cleanup func() error, err error) {
	envCleanup, err := superdTesting.SetupEnv()
	if err != nil {
		return
	}

	cwd, err := os.Getwd()
	if err != nil {
		return
	}

	testDataDir := filepath.Clean(filepath.Join(cwd, "../../test"))

	// copy test services to XDG_CONFIG_HOME
	serviceDir := os.ExpandEnv("$XDG_CONFIG_HOME/services")
	if err = os.MkdirAll(serviceDir, 0750); err != nil {
		return
	}

	serviceFiles, err := filepath.Glob(filepath.Join(testDataDir, "*.service"))
	if err != nil {
		return
	}

	for _, f := range serviceFiles {
		if err = copyFile(f, filepath.Join(serviceDir, filepath.Base(f))); err != nil {
			return
		}
	}

	// start logger
	logFilePath := os.ExpandEnv("$XDG_STATE_HOME/superd.log")
	logger, loggingStop, err := startLogging(logFilePath, true)
	if err != nil {
		panic(err)
	}
	logger.SetMinLevel(log.DEBUG)

	// start manager
	mgr = manager.Init("", false)
	mgrStop := make(chan struct{})
	mgrReady := make(chan struct{})
	go mgr.Start(mgrStop, mgrReady)

	select {
	case <-time.After(2 * time.Second):
		err = fmt.Errorf("timed out waiting for the manager to become ready")
		return
	case <-mgrReady:
		break
	}

	cleanup = func() (err error) {
		close(mgrStop)

		if err = loggingStop(); err != nil {
			return
		}

		if err = envCleanup(); err != nil {
			return
		}

		return
	}

	return
}

// Get the state for the given service, and compare it to the expectedState.
// This will return an error if the state is unexpected, more than 1 state is
// returned from the manager, or there is some other error from the manager.
func verifyState(service string, expectedState api.ServiceState) error {
	if status, err := mgr.Status([]string{service}); err != nil {
		return fmt.Errorf("unexpected error received: %w", err)
	} else if len(status) != 1 {
		return fmt.Errorf("unexpected number of status returned, expected 1, got %d", len(status))
	} else if status[service].State != expectedState {
		return fmt.Errorf("expected state: %s, got: %s", expectedState, status[service].State)
	}

	return nil
}

func TestMain(m *testing.M) {
	// these tests take a little longer to run, so allow skipping them
	if os.Getenv("SUPERD_SKIP_INTEGRATION") != "" {
		fmt.Println("SUPERD_SKIP_INTEGRATION is set, skipping integration tests...")
		return
	}

	cleanup, err := setup()
	if err != nil {
		panic(err)
	}

	ret := m.Run()

	defer func() {

		if err := cleanup(); err != nil {
			panic(err)
		}
	}()

	os.Exit(ret)
}

func TestServiceControl(t *testing.T) {

	type action struct {
		// 'start' 'stop'
		action        string
		expectedState api.ServiceState
		// really simple error verification. TODO: add/use custom errors?
		expectedErr bool
	}

	subtests := []struct {
		name string
		in   []action
	}{
		{
			name: "testoneshot_success",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceStarted,
				},
				{
					action:        "stop",
					expectedState: api.ServiceStopped,
				},
			},
		},
		{
			name: "testoneshot_fail",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceFailed,
					expectedErr:   true,
				},
				{
					action:        "stop",
					expectedState: api.ServiceFailed,
				},
			},
		},
		{
			name: "testoneshot_nonexistent",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceFailed,
					expectedErr:   true,
				},
				{
					action:        "stop",
					expectedState: api.ServiceFailed,
				},
			},
		},
		{
			name: "testsimple_success",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceStarted,
				},
				{
					action:        "stop",
					expectedState: api.ServiceStopped,
				},
			},
		},
		{
			// see todo#37
			name: "testforkedchild",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceStarted,
				},
				{
					action:        "stop",
					expectedState: api.ServiceStopped,
				},
			},
		},
		{
			name: "testdependency_success",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceStarted,
				},
				{
					action:        "stop",
					expectedState: api.ServiceStopped,
				},
			},
		},
		{
			name: "testdependency_fail",
			in: []action{
				{
					action:        "start",
					expectedState: api.ServiceStopped,
					expectedErr:   true,
				},
				{
					action:        "stop",
					expectedState: api.ServiceStopped,
				},
			},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			// these tests take a long time (hundreds of ms each on a fast
			// laptop), explicitly allowing them to run in parallel speeds
			// things up considerably
			t.Parallel()

			for _, a := range subtest.in {
				var err error = nil
				switch a.action {
				case "start":
					err = mgr.StartService(subtest.name)
				case "stop":
					err = mgr.StopService(subtest.name)
				case "restart":
					err = mgr.RestartService(subtest.name)
				default:
					t.Fatalf("unknown action to test: %s", a.action)
				}

				if err != nil {
					if !a.expectedErr {
						t.Fatalf("unexpected failure when trying to %s service %q: %s", a.action, subtest.name, err)
					}
				} else if a.expectedErr {
					t.Fatal("expected error but none received")
				}

				start := time.Now()
				for {
					if err = verifyState(subtest.name, a.expectedState); err == nil {
						break
					}
					if time.Since(start) > 500*time.Millisecond {
						// timed out looking for expected state, so assume
						// whatever err is set is truly valid
						break
					}
				}
				if err != nil {
					t.Fatalf("when doing %q: %s", a.action, err)
				}
			}
		})
	}
}
