// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package cmd

import (
	"fmt"
	"net"
	"os"
	"path/filepath"
	"sync"
	"syscall"
	"time"

	goCmd "git.sr.ht/~craftyguy/go-cmd"
)

type ErrTimeout struct {
	Cmd string
}

func (e ErrTimeout) Error() string {
	return fmt.Sprint("timed out running command: ", e.Cmd)
}

type ErrCmdExitedNonZero struct {
	Ret int
}

func (e ErrCmdExitedNonZero) Error() string {
	return fmt.Sprint("command exited with non-zero return code: ", e.Ret)
}

type ErrCmdExited struct{}

func (e ErrCmdExited) Error() string {
	return "command exited"
}

type ErrUnhandledStatus struct {
	Msg string
}

func (e ErrUnhandledStatus) Error() string {
	return fmt.Sprint("unhandled status received: ", e.Msg)
}

var (
	// for mocking in tests
	execCommand = goCmd.NewCmdOptions
	// for mocking in tests
	sockHandler = notifySockHandler
	// give 100ms for process to start
	startTimerDuration = 100 * time.Millisecond
	upCheck            = isUp
)

type Streams struct {
	// TODO: wrap goCmd.Status
	Status <-chan goCmd.Status
	Stdout <-chan string
	Stderr <-chan string
}

type Commander interface {
	Status() goCmd.Status
	Start(env []string) (err error)
	StartNotify(env []string, sockPath string, startTimeoutDuration time.Duration) (err error)
	Stop(timeout time.Duration) error
	Streams() *Streams
}

type Command struct {
	cmd     *goCmd.Cmd
	streams *Streams
	sync.Mutex
}

func NewCommand(exe string, args []string, out *os.File) Commander {
	c := &Command{
		cmd: execCommand(goCmd.Options{
			Streaming: false,
			Buffered:  false,
			File:      out,
		}, exe, args...),
	}
	return c
}

func (c *Command) Status() goCmd.Status {
	return c.cmd.Status()
}

func (c *Command) Stop(timeout time.Duration) error {
	go func() {
		c.Lock()
		defer c.Unlock()

		if c.cmd.Status().Complete {
			return
		}

		c.cmd.Stop()
		for {
			select {
			case <-time.After(timeout):
				// failed to stop with SIGTERM, so use SIGKILL
				c.cmd.SendSignal(syscall.SIGKILL, true)
				// if the process stays running after a SIGKILL, there's
				// nothing else we can do so just bail
				return
			default:
				if c.cmd.Status().StopTs > 0 {
					return
				}
				time.Sleep(1 * time.Millisecond)
			}
		}
	}()
	return nil
}

func (c *Command) Streams() *Streams {
	return c.streams
}

// Runs the command, returning the status and the details/properties of the
// started process if it was successful.
func (c *Command) Start(env []string) (err error) {
	var startTimer <-chan time.Time

	c.cmd.Env = env

	c.streams = &Streams{
		Status: c.cmd.Start(),
		Stdout: c.cmd.Stdout,
		Stderr: c.cmd.Stderr,
	}

	startTimer = time.After(startTimerDuration)

	select {
	// Start timer
	case <-startTimer:
		// start timer expired, if exe is still up then it's probably fine
		startTimer = nil

		if !upCheck(c.Status().PID) {
			err = ErrTimeout{c.cmd.Name}
			return
		}

	case pStatus := <-c.streams.Status:
		// status from go-cmd
		if pStatus.Complete {
			if pStatus.Exit != 0 {
				err = ErrCmdExitedNonZero{pStatus.Exit}
			} else {
				err = ErrCmdExited{}
			}
			return
		} else {
			err = ErrUnhandledStatus{pStatus.Error.Error()}
		}
	}

	return
}

// socket connection handler that closes notify chan when READY=1 is received
// TODO: obviously this needs to be refactored to support other messages, but
// it's doubtful I/superd will care about other things can come over the notify
// socket, so until we care, this simpler implementation is good enough.
func notifySockHandler(socketPath string) (notify chan struct{}, err error) {
	notifySocket, err := net.Listen("unix", socketPath)
	if err != nil {
		err = fmt.Errorf("unable to listen on notify socket: %w", err)
		return
	}

	notify = make(chan struct{})

	go func() {
		var conn net.Conn
		for {
			conn, err = (notifySocket).Accept()
			if err != nil {
				break
			}
			buf := make([]byte, 64)
			_, err = (conn).Read(buf)
			if err != nil {
				break
			}
			if string(buf) == "READY=1" {
				close(notify)
				break
			}
		}
		notifySocket.Close()
	}()

	return
}

func (c *Command) StartNotify(env []string, sock string, startTimeoutDuration time.Duration) (err error) {

	// create a notify socket and pass path to service using NOTIFY_SOCKET
	runtimeDir, ok := os.LookupEnv("XDG_RUNTIME_DIR")
	if !ok {
		return fmt.Errorf("XDG_RUNTIME_DIR is unset, but required for Type=notify")
	}

	socketPath := filepath.Join(runtimeDir, sock)
	if err := os.RemoveAll(socketPath); err != nil {
		return fmt.Errorf("unable to remove old notify socket: %w", err)
	}

	c.cmd.Env = env
	c.cmd.Env = append(c.cmd.Env, "NOTIFY_SOCKET="+socketPath)

	c.streams = &Streams{
		Status: c.cmd.Start(),
		Stdout: c.cmd.Stdout,
		Stderr: c.cmd.Stderr,
	}

	notifyCh, err := sockHandler(socketPath)
	if err != nil {
		return err
	}

	select {
	// Start timer
	case <-time.After(startTimeoutDuration):
		err = ErrTimeout{c.cmd.Name}
	case <-notifyCh:
		// service has started and notified us
	}

	return
}

func isUp(pid int) bool {
	up := true

	p, err := os.FindProcess(pid)
	if err != nil {
		up = false
	} else {
		if err := p.Signal(syscall.Signal(0)); err != nil {
			up = false
		}
	}

	return up
}
