// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package cmd

import (
	"os"
	"path/filepath"
	"reflect"
	"strconv"
	"testing"
	"time"

	goCmd "git.sr.ht/~craftyguy/go-cmd"
	superdTesting "sr.ht/~craftyguy/superd/internal/testing"
)

func TestMain(m *testing.M) {
	if os.Getenv("TEST_MAIN") == "start" {
		StartExec()
		return
	}

	cleanup, err := superdTesting.SetupEnv()
	if err != nil {
		panic(err)
	}

	ret := m.Run()

	if err := cleanup(); err != nil {
		panic(err)
	}

	os.Exit(ret)
}

func StartExec() {
	ms, _ := strconv.Atoi(os.Getenv("RUNTIME_MS"))
	time.Sleep(time.Duration(ms) * time.Millisecond)

	ret, _ := strconv.Atoi(os.Getenv("EXIT_CODE"))
	os.Exit(ret)
}

func TestStart(t *testing.T) {
	subtests := []struct {
		name            string
		inRet           int
		inUpCheckRet    bool
		inRuntimeMs     int
		inStartTimerDur time.Duration
		expectedErr     error
	}{
		{
			name:            "started no failure",
			inUpCheckRet:    true,
			inRuntimeMs:     2000,
			inStartTimerDur: 100 * time.Millisecond,
			expectedErr:     nil,
		},
		{
			name:  "return non-zero",
			inRet: 4,
			// made really long so it doesn't 'normally' time out, which might
			// happen if it takes the OS/test system too long to actually run
			// the test exe
			inStartTimerDur: 5 * time.Second,
			expectedErr:     ErrCmdExitedNonZero{},
		},
		{
			name:  "return zero",
			inRet: 0,
			// made really long so it doesn't 'normally' time out, which might
			// happen if it takes the OS/test system too long to actually run
			// the test exe
			inStartTimerDur: 5 * time.Second,
			expectedErr:     ErrCmdExited{},
		},
		{
			name:            "timeout",
			inUpCheckRet:    false,
			inRuntimeMs:     2000,
			inStartTimerDur: 100 * time.Millisecond,
			expectedErr:     ErrTimeout{},
		},
	}

	var upCheckRet bool
	upCheck = func(pid int) bool {
		if pid == 0 {
			t.Fatalf("received unexpected pid: %d", pid)
		}

		return upCheckRet
	}
	defer func() {
		upCheck = isUp
		execCommand = goCmd.NewCmdOptions
		startTimerDuration = 100 * time.Millisecond
	}()

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			startTimerDuration = st.inStartTimerDur

			upCheckRet = st.inUpCheckRet

			execCommand = func(options goCmd.Options, name string, args ...string) *goCmd.Cmd {
				cmd := goCmd.NewCmdOptions(options, os.Args[0])
				cmd.Args = args
				return cmd
			}

			c := NewCommand("foo", []string{}, nil)
			env := []string{
				"TEST_MAIN=start",
				"EXIT_CODE=" + strconv.Itoa(st.inRet),
				"RUNTIME_MS=" + strconv.Itoa(st.inRuntimeMs),
			}
			if err := c.Start(env); err != nil {
				// note: TypeOf used instead of errors.Is since an instance of
				// the error is stored in the subtest (expectedErr), and I'm
				// unaware of any way to "store a type" for the comparison here
				if reflect.TypeOf(err) != reflect.TypeOf(st.expectedErr) {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			} else {
				if st.expectedErr != nil {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			}
		})
	}
}

func TestStartNotify(t *testing.T) {
	subtests := []struct {
		name           string
		inNotify       bool
		inRet          int
		inRuntimeMs    int
		inStartTimeout time.Duration
		inSockName     string
		expectedErr    error
	}{
		{
			name:           "started no failure",
			inNotify:       true,
			inStartTimeout: 1 * time.Second,
			inSockName:     "foo.sock",
			expectedErr:    nil,
		},
		{
			name:     "no notify, timeout",
			inNotify: false,
			// timeout quickly so the test doesn't take too long
			inStartTimeout: 1 * time.Millisecond,
			inSockName:     "foo.sock",
			expectedErr:    ErrTimeout{},
		},
	}

	defer func() {
		execCommand = goCmd.NewCmdOptions
		sockHandler = notifySockHandler
	}()

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {

			execCommand = func(options goCmd.Options, name string, args ...string) *goCmd.Cmd {
				cmd := goCmd.NewCmdOptions(options, os.Args[0])
				cmd.Args = args
				return cmd
			}

			c := NewCommand("foo", []string{}, nil)
			// cmd.go currently puts the socket in XDG_RUNTIME_DIR
			expectedSockPath := filepath.Join(os.Getenv("XDG_RUNTIME_DIR"), "foo.sock")
			env := []string{
				"TEST_MAIN=startNotify",
				"EXIT_CODE=" + strconv.Itoa(st.inRet),
				"RUNTIME_MS=" + strconv.Itoa(st.inRuntimeMs),
				"EXPECTED_NOTIFY=" + strconv.FormatBool(st.inNotify),
			}

			// mocked socket connection handler, so actual sockets aren't created/used
			sockHandler = func(socketPath string) (notify chan struct{}, err error) {
				if socketPath != expectedSockPath {
					t.Fatalf("expected: %s, received: %s", expectedSockPath, socketPath)
				}

				notify = make(chan struct{})

				if st.inNotify {
					close(notify)
				}

				return
			}

			if err := c.StartNotify(env, st.inSockName, st.inStartTimeout); err != nil {
				// note: TypeOf used instead of errors.Is since an instance of
				// the error is stored in the subtest (expectedErr), and I'm
				// unaware of any way to "store a type" for the comparison here
				if reflect.TypeOf(err) != reflect.TypeOf(st.expectedErr) {
					t.Log(err)
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			} else {
				if st.expectedErr != nil {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			}
		})
	}
}
