// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package config

import (
	"bytes"
	"fmt"
	"os"
	"path/filepath"
	"regexp"
	"strconv"
	"strings"
	"time"

	ini "git.sr.ht/~spc/go-ini"
	"github.com/google/shlex"
	"github.com/trstringer/go-systemd-time/pkg/systemdtime"
)

type unit struct {
	Description           string
	After                 string
	Before                string
	StartLimitIntervalSec string
	StartLimitBurst       string
}

type service struct {
	Type            string
	Restart         string
	RestartSec      string
	ExecStart       []string
	ExecStartPre    []string
	ExecStartPost   []string
	ExecOnFailure   []string
	TimeoutStartSec string
	Environment     string
}

type systemdConfig struct {
	Unit    unit
	Service service
}

type ErrUnsupportedValue struct {
	Option string
	Value  string
}

func (e ErrUnsupportedValue) Error() string {
	return fmt.Sprintf("unsupported value for %q: %s", e.Option, e.Value)
}

type ErrInvalidField struct {
	Field   string
	Message string
}

func (e ErrInvalidField) Error() string {
	return fmt.Sprintf("unable to parse field %q: %s", e.Field, e.Message)
}

const (
	// copied from systemd defaults
	defaultRestartDur      = 100 * time.Millisecond
	defaultStartLimitDur   = 10 * time.Second
	defaultStartLimitBurst = 5
	defaultTimeoutStartDur = 90 * time.Second
)

func replaceInstanceName(contents []byte, name string) (out []byte) {
	out = bytes.ReplaceAll(contents, []byte("%i"), []byte(name))
	return
}

func multiline2single(s string) string {
	out := strings.TrimSpace(s)
	out = strings.ReplaceAll(out, "\n", "")
	out = strings.ReplaceAll(out, "\r", "")
	return strings.ReplaceAll(out, "\\", "")
}

// Parse the "unit" name from the path.
// Expects the path to be to a 'full' instanced template file, e.g.
// /path/to/foo@bar.service and *not* /path/to/foo@.service
// Returns the instance name, or an empty string if unable to find it
func instanceName(path string) (name string) {
	file := filepath.Base(strings.TrimSuffix(path, ".service"))
	if strings.HasPrefix(file, "@") {
		return
	}

	parts := strings.Split(file, "@")
	if len(parts) != 2 {
		return
	}

	name = parts[1]
	return
}

func Parse(path string) (s *ServiceConfig, err error) {
	// TODO: this ignore escaping of "unit names"... so %i is basically %I
	// TODO: also implement other specifiers?
	// see: https://www.freedesktop.org/software/systemd/man/systemd.unit.html

	template, err := regexp.MatchString(".+@.+.service", filepath.Base(path))
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	contents, err := os.ReadFile(path)
	if err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	if template {
		contents = replaceInstanceName(contents, instanceName(path))
	}

	sysd := &systemdConfig{}

	o := ini.Options{
		AllowNumberSignComments: true,
		AllowMultilineValues:    true,
		AllowEmptyValues:        true,
	}

	if err = ini.UnmarshalWithOptions(contents, sysd, o); err != nil {
		err = fmt.Errorf("config.Parse: %w", err)
		return
	}

	name := strings.Split(filepath.Base(path), filepath.Ext(path))[0]

	s, err = parseConfig(name, sysd)
	if err != nil {
		return
	}

	s.Path = path

	return
}

func parseDeps(s string) (deps []string, err error) {

	for _, dep := range strings.Fields(strings.TrimSpace(s)) {
		if dep == "" {
			continue
		}
		// Sometimes dependencies are listed with a ".service" extension, those
		// are allowed here, but any other is an error since we don't want to
		// allow things that aren't services (e.g. a ".socket", if that's even
		// valid..)
		if ext := filepath.Ext(dep); ext != "" {
			if strings.ToLower(ext) == ".service" {
				dep = strings.TrimSuffix(dep, ext)
			} else {
				err = ErrUnsupportedValue{Option: "Before=|After=", Value: s}
				fmt.Println(err)
				return
			}
		}
		deps = append(deps, dep)
	}
	return
}

func parseCmd(lines []string, allowEmpty bool) (execs []Exec, err error) {

	if len(lines) < 1 {
		if !allowEmpty {
			err = ErrUnsupportedValue{Option: "ExecStart*=", Value: ""}
		}
		return
	}

	for _, line := range lines {
		if line == "" {
			if !allowEmpty {
				err = ErrUnsupportedValue{Option: "ExecStart*=", Value: ""}
				return
			}
			continue
		}

		// // split ExecStart into cmd string + args slice
		var (
			cmd    string
			args   []string
			tokens []string
		)

		// TODO: shlex still cannot handle nested double quotes, e.g: /bin/sh -c "echo \"hello world\""
		tokens, err = shlex.Split(multiline2single(line))
		if err != nil {
			err = ErrUnsupportedValue{Option: "ExecStart*=", Value: err.Error()}
			return
		}

		if len(tokens) < 1 {
			err = ErrUnsupportedValue{Option: "ExecStart*=", Value: strings.Join(tokens, " ")}
			return
		}

		cmd = tokens[0]
		if len(tokens) > 1 {
			args = tokens[1:]
		}

		// expand any env vars in the arguments. this loosely mimicks systemd's
		// "basic" variable expansion support, and probably needs to be revisited
		// at some point. TODO!
		for i := range args {
			args[i] = os.ExpandEnv(args[i])
		}

		var allowedToFail bool
		if strings.HasPrefix(cmd, "-") {
			cmd = strings.TrimPrefix(cmd, "-")
			allowedToFail = true
		}
		execs = append(execs, Exec{
			Cmd:           cmd,
			Args:          args,
			AllowedToFail: allowedToFail,
		})
	}

	return
}

func parseRestart(s string) (restart RestartCond, err error) {

	switch strings.ToUpper(strings.TrimSpace(s)) {
	case "ALWAYS":
		restart = RestartAlways
	case "ON-FAILURE":
		restart = RestartFailure
	case "ON-SUCCESS":
		restart = RestartSuccess
	case "":
		fallthrough
	case "NO":
		restart = RestartNever
	default:
		err = ErrUnsupportedValue{Option: "Restart=", Value: s}
	}

	return
}

func parseDur(s string, optionName string, required bool, defaultDur time.Duration) (dur time.Duration, err error) {

	if required && s == "" {
		err = ErrUnsupportedValue{Option: optionName, Value: "value required but none was given"}
		return
	}

	if s != "" {
		if dur, err = systemdtime.ToDuration(strings.TrimSpace(s)); err != nil {
			// assume it's in seconds and try again
			if dur, err = systemdtime.ToDuration(s + "s"); err != nil {
				err = ErrUnsupportedValue{Option: optionName, Value: s}
			}
		}
	} else {
		dur = defaultDur
	}

	return
}

func parseType(s string) (t Type, err error) {
	switch strings.ToUpper(strings.TrimSpace(s)) {
	case "":
		fallthrough
	case "SIMPLE":
		t = TypeSimple
	case "ONESHOT":
		t = TypeOneshot
	case "NOTIFY":
		t = TypeNotify
	default:
		err = ErrUnsupportedValue{Option: "Type=", Value: s}
	}

	return
}

// parse the given string as an int, if the string is empty then defaultVal is
// returned. if the string cannot be converted to an int, an
// ErrUnsupportedValue is returned
func parseInt(s string, defaultVal int) (i int, err error) {
	v := multiline2single(s)
	if v == "" {
		return defaultVal, nil
	}
	ret, err := strconv.Atoi(v)
	if err != nil {
		err = ErrUnsupportedValue{Option: "Restart=", Value: s}
	}

	return ret, err
}

// parse environment variable/value pairs from the given string
func parseEnv(s string) (env map[string]string, err error) {
	env = make(map[string]string)
	groups := []string{}
	group := ""
	quoted := false
	for _, ch := range multiline2single(s) {
		if ch == '"' || ch == '\'' {
			if quoted {
				quoted = false
			} else {
				quoted = true
			}
			continue
		}
		if ch == ' ' && !quoted {
			groups = append(groups, group)
			group = ""
		} else if ch == '\n' {
			groups = append(groups, group)
		} else {
			group += string(ch)
		}
	}
	groups = append(groups, group)

	for _, group := range groups {
		before, after, found := strings.Cut(group, "=")
		if !found {
			continue
		}
		env[before] = after
	}

	return
}

func parseConfig(name string, sysd *systemdConfig) (s *ServiceConfig, err error) {
	s = &ServiceConfig{
		Name:       name,
		RestartDur: defaultRestartDur,
	}

	// Description=
	s.Description = multiline2single(sysd.Unit.Description)

	// "Before" dependencies
	if s.Before, err = parseDeps(sysd.Unit.Before); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// "After" dependencies
	if s.After, err = parseDeps(sysd.Unit.After); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Type=
	// Note: this needs to be set *before* ExecStart, so that some checks can
	// be enforced when parsing that parameter
	if s.Type, err = parseType(sysd.Service.Type); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecStart=
	s.ExecStart, err = parseCmd(sysd.Service.ExecStart, false)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}
	if s.Type != TypeOneshot {
		if len(s.ExecStart) != 1 {
			err = ErrInvalidField{Field: "ExecStart=",
				Message: "multiple ExecStart= are only supported for Type=oneshot"}
			return
		} else if s.ExecStart[0].AllowedToFail {
			err = ErrInvalidField{Field: "ExecStart=",
				Message: "cannot be allowed to fail (has a '-' prefix)"}
			return
		}
	}

	// StartLimitIntervalSec=
	if s.StartLimitDur, err = parseDur(sysd.Unit.StartLimitIntervalSec, "StartLimitIntervalSec", false, defaultStartLimitDur); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// StartLimitBurst=
	if s.StartLimitBurst, err = parseInt(sysd.Unit.StartLimitBurst, 0); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}
	if s.StartLimitBurst == 0 {
		s.StartLimitBurst = defaultStartLimitBurst
	}

	// ExecStartPre=
	s.ExecStartPre, err = parseCmd(sysd.Service.ExecStartPre, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// // ExecStartPost=
	s.ExecStartPost, err = parseCmd(sysd.Service.ExecStartPost, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// ExecOnFailure=
	s.ExecOnFailure, err = parseCmd(sysd.Service.ExecOnFailure, true)
	if err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Restart=
	if s.Restart, err = parseRestart(sysd.Service.Restart); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// RestartSec=
	if s.RestartDur, err = parseDur(sysd.Service.RestartSec, "Restart=", false, defaultRestartDur); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// TimeoutStartSec=
	if s.TimeoutStartDur, err = parseDur(sysd.Service.TimeoutStartSec, "TimeoutStartSec", false, defaultTimeoutStartDur); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	// Environment=
	if s.Env, err = parseEnv(sysd.Service.Environment); err != nil {
		err = fmt.Errorf("%s: %w", s.Name, err)
		return
	}

	return
}
