// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package config

import (
	"bytes"
	"encoding/json"
	"reflect"
	"testing"
	"time"
)

func TestReplaceInstanceName(t *testing.T) {
	subtests := []struct {
		name     string
		inBytes  []byte
		inName   string
		expected []byte
	}{
		{
			name:     "only 1 %i",
			inBytes:  []byte{69, 120, 101, 99, 61, 98, 97, 114, 32, 45, 99, 32, 37, 105, 32, 45, 118, 32, 45, 45, 119, 104, 97, 116, 101, 118, 101, 114},
			inName:   "fooooooooo",
			expected: []byte{69, 120, 101, 99, 61, 98, 97, 114, 32, 45, 99, 32, 102, 111, 111, 111, 111, 111, 111, 111, 111, 111, 32, 45, 118, 32, 45, 45, 119, 104, 97, 116, 101, 118, 101, 114},
		},
		{
			name:     "no %i",
			inBytes:  []byte{37, 32, 105, 32, 37, 98, 32, 105, 37},
			inName:   "pretzels",
			expected: []byte{37, 32, 105, 32, 37, 98, 32, 105, 37},
		},
		{
			name:     "3 %i's",
			inBytes:  []byte{69, 120, 101, 99, 83, 116, 97, 114, 116, 61, 100, 111, 115, 116, 117, 102, 102, 32, 37, 105, 32, 45, 99, 32, 37, 105, 32, 37, 105, 32, 37, 106, 32, 37, 107, 32, 37, 108},
			inName:   "twix",
			expected: []byte{69, 120, 101, 99, 83, 116, 97, 114, 116, 61, 100, 111, 115, 116, 117, 102, 102, 32, 116, 119, 105, 120, 32, 45, 99, 32, 116, 119, 105, 120, 32, 116, 119, 105, 120, 32, 37, 106, 32, 37, 107, 32, 37, 108},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out := replaceInstanceName(subtest.inBytes, subtest.inName)
			if !bytes.Equal(out, subtest.expected) {
				t.Fatalf("Expected: %q, got: %q", subtest.expected, out)
			}
		})
	}
}

func TestInstanceName(t *testing.T) {
	subtests := []struct {
		name     string
		in       string
		expected string
	}{
		{
			name:     "basic instance name",
			in:       "/path/to/some/foo@bar.service",
			expected: "bar",
		},
		{
			name:     "template",
			in:       "/etc/services/@bar.service",
			expected: "",
		},
		{
			name:     "valid instance with deep path",
			in:       "/////////////foo@bar.service",
			expected: "bar",
		},
		{
			name:     "invalid template/instance name",
			in:       "/path/to/some/foo@.service",
			expected: "",
		},
		{
			name:     "non-template config file",
			in:       "/path/to/some/foo.service",
			expected: "",
		},
		{
			name:     "not a valid anything",
			in:       "apples?!",
			expected: "",
		},
		{
			name:     "file with many @s",
			in:       "foo@bar@bazz.service",
			expected: "",
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out := instanceName(subtest.in)
			if out != subtest.expected {
				t.Fatalf("Expected: %q, got: %q", subtest.expected, out)
			}
		})
	}
}

func TestParseConfig(t *testing.T) {
	subtests := []struct {
		name        string
		in          systemdConfig
		expected    *ServiceConfig
		expectedErr error
	}{
		{
			name: "valid config",
			in: systemdConfig{
				Unit: unit{
					Description:           "this is a test of a valid .service config file",
					After:                 "bazz boop",
					Before:                "bar",
					StartLimitIntervalSec: "3s",
					StartLimitBurst:       "12",
				},
				Service: service{
					Type:            "Simple",
					ExecStart:       []string{"/usr/bin/foo --bar --bazz 127"},
					ExecStartPre:    []string{},
					ExecStartPost:   []string{"/bin/true"},
					ExecOnFailure:   []string{"/bin/alert -b everybody_panic"},
					Restart:         "Always",
					RestartSec:      "4",
					TimeoutStartSec: "1337s",
					Environment:     "BANANAS=nope \"FOO=dragon fruit\"",
				},
			},
			expected: &ServiceConfig{
				Name:        "validconfigtest",
				Description: "this is a test of a valid .service config file",
				Type:        TypeSimple,
				Restart:     RestartAlways,
				RestartDur:  4 * time.Second,
				ExecStart: []Exec{
					{
						Cmd:  "/usr/bin/foo",
						Args: []string{"--bar", "--bazz", "127"},
					},
				},
				ExecStartPost: []Exec{
					{
						Cmd: "/bin/true",
					},
				},
				ExecOnFailure: []Exec{
					{
						Cmd:  "/bin/alert",
						Args: []string{"-b", "everybody_panic"},
					},
				},
				Before:          []string{"bar"},
				After:           []string{"bazz", "boop"},
				StartLimitDur:   3 * time.Second,
				StartLimitBurst: 12,
				TimeoutStartDur: 1337 * time.Second,
				Env: map[string]string{
					"BANANAS": "nope",
					"FOO":     "dragon fruit",
				},
			},
			expectedErr: nil,
		},
		{
			name: "multiple ExecStartPre",
			in: systemdConfig{
				Unit: unit{
					Description:           "multiple ExecStartPre",
					StartLimitIntervalSec: "3s",
					StartLimitBurst:       "12",
				},
				Service: service{
					Type:      "Simple",
					ExecStart: []string{"/usr/bin/foo --bar --bazz 127"},
					ExecStartPre: []string{
						"mkdir /tmp/bazz",
						"rm -r /tmp/bazz",
					},
					TimeoutStartSec: "1337s",
					Restart:         "Always",
					RestartSec:      "4",
				},
			},
			expected: &ServiceConfig{
				Name:        "multiple_execstartpre_test",
				Description: "multiple ExecStartPre",
				Type:        TypeSimple,
				ExecStart: []Exec{
					{
						Cmd:  "/usr/bin/foo",
						Args: []string{"--bar", "--bazz", "127"},
					},
				},
				ExecStartPre: []Exec{
					{
						Cmd:  "mkdir",
						Args: []string{"/tmp/bazz"},
					},
					{
						Cmd:  "rm",
						Args: []string{"-r", "/tmp/bazz"},
					},
				},
				StartLimitDur:   3 * time.Second,
				StartLimitBurst: 12,
				TimeoutStartDur: 1337 * time.Second,
				Restart:         RestartAlways,
				RestartDur:      4 * time.Second,
				Env:             map[string]string{},
			},
			expectedErr: nil,
		},
		{
			name: "multiple ExecStart, not 'oneshot'",
			in: systemdConfig{
				Unit: unit{
					Description: "multiple ExecStart, not 'oneshot'",
				},
				Service: service{
					Type: "Simple",
					ExecStart: []string{
						"/usr/bin/foo --bar --bazz 127",
						"/usr/bin/superd -v ",
					},
				},
			},
			expected: &ServiceConfig{
				Name:        "multiple_execstart_not_oneshot",
				Description: "multiple ExecStart, not 'oneshot'",
				Type:        TypeSimple,
			},
			expectedErr: ErrInvalidField{},
		},
		{
			name: "ExecStart, with - prefix",
			in: systemdConfig{
				Unit: unit{
					Description: "ExecStart, with - prefix",
				},
				Service: service{
					Type:      "Simple",
					ExecStart: []string{"-/usr/bin/foo --bar --bazz 127"},
				},
			},
			expected: &ServiceConfig{
				Name:        "execstart_with_-_prefix",
				Description: "ExecStart, with - prefix",
				Type:        TypeSimple,
			},
			expectedErr: ErrInvalidField{},
		},
		{
			name: "multiple ExecStart, type=oneshot",
			in: systemdConfig{
				Unit: unit{
					Description:           "multiple ExecStart, type=oneshot",
					StartLimitIntervalSec: "3s",
					StartLimitBurst:       "12",
				},
				Service: service{
					Type: "oneshot",
					ExecStart: []string{
						"/usr/bin/foo --bar --bazz 127",
						"/usr/bin/superd -v ",
					},
					ExecStartPre: []string{
						"-mkdir /tmp/bazz",
						"rm -r /tmp/bazz",
					},
					TimeoutStartSec: "1337s",
					Restart:         "Always",
					RestartSec:      "4",
				},
			},
			expected: &ServiceConfig{
				Name:        "multiple_execstart_oneshot",
				Description: "multiple ExecStart, type=oneshot",
				Type:        TypeOneshot,
				ExecStart: []Exec{
					{
						Cmd:  "/usr/bin/foo",
						Args: []string{"--bar", "--bazz", "127"},
					},
					{
						Cmd:  "/usr/bin/superd",
						Args: []string{"-v"},
					},
				},
				ExecStartPre: []Exec{
					{
						Cmd:           "mkdir",
						Args:          []string{"/tmp/bazz"},
						AllowedToFail: true,
					},
					{
						Cmd:  "rm",
						Args: []string{"-r", "/tmp/bazz"},
					},
				},
				StartLimitDur:   3 * time.Second,
				StartLimitBurst: 12,
				TimeoutStartDur: 1337 * time.Second,
				Restart:         RestartAlways,
				RestartDur:      4 * time.Second,
				Env:             map[string]string{},
			},
			expectedErr: nil,
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseConfig(subtest.expected.Name, &subtest.in)
			if err != nil {
				// note: TypeOf used instead of errors.Is since an instance of
				// the error is stored in the subtest (expectedErr), and I'm
				// unaware of any way to "store a type" for the comparison here
				if reflect.TypeOf(err) != reflect.TypeOf(subtest.expectedErr) {
					t.Fatalf("err expected: %T, err received: %T", subtest.expectedErr, err)
				}
				// this was the test, congrats you passed
				return
			}
			if !reflect.DeepEqual(out, subtest.expected) {
				o, _ := json.MarshalIndent(out, "", "\t")
				e, _ := json.MarshalIndent(subtest.expected, "", "\t")
				t.Fatalf("\n**Expected: %s\n**Got: %s", string(e), string(o))
			}
		})
	}
}

func TestParseRestartDur(t *testing.T) {
	subtests := []struct {
		name string
		in   string
		// string to feed directly to time.ParseDuration
		expected    string
		expectedErr bool
	}{
		{
			name:     "int value as seconds",
			in:       "9000",
			expected: "9000s",
		},
		{
			name:     "time span in systemd.time format",
			in:       "6h 20m 11s",
			expected: "6h20m11s",
		},
		{
			name:     "valid but backwards",
			in:       "1s 10m",
			expected: "10m1s",
		},
		{
			name:        "invalid value",
			in:          "banananananananananana",
			expectedErr: true,
		},
		{
			name:     "int value as seconds, with newline",
			in:       "9000\r\n",
			expected: "9000s",
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseDur(subtest.in, "Restart=", false, defaultRestartDur)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			e, err := time.ParseDuration(subtest.expected)
			if err != nil {
				t.Fatalf("unable to parse expected time duration of %q: %s", subtest.expected, err)
			}

			if out != e {
				t.Fatalf("expected: %q, got: %q", e, out)
			}
		})
	}
}

func TestParseCmd(t *testing.T) {
	subtests := []struct {
		name         string
		in           []string
		inAllowEmpty bool
		expected     []Exec
		expectedErr  bool
	}{
		{
			name: "valid command + args",
			in:   []string{"/usr/bin/flup -a -T 32"},
			expected: []Exec{
				{
					Cmd:  "/usr/bin/flup",
					Args: []string{"-a", "-T", "32"},
				},
			},
		},
		{
			name: "valid command + no args",
			in:   []string{"sway"},
			expected: []Exec{
				{
					Cmd:  "sway",
					Args: nil,
				},
			},
		},
		{
			name:        "empty and allowEmpty=false",
			in:          []string{},
			expectedErr: true,
		},
		{
			name:         "empty and allowEmpty=true",
			in:           []string{},
			inAllowEmpty: true,
		},
		{
			name: "valid command, multiline",
			in:   []string{"sleep \\ 123 \\ -h"},
			expected: []Exec{
				{
					Cmd:  "sleep",
					Args: []string{"123", "-h"},
				},
			},
		},
		{
			name: "multiple commands",
			in: []string{
				"sleep \\ 123 \\ -h",
				"foo --bar --bananananana",
			},
			expected: []Exec{
				{
					Cmd:  "sleep",
					Args: []string{"123", "-h"},
				},
				{
					Cmd:  "foo",
					Args: []string{"--bar", "--bananananana"},
				},
			},
		},
		{
			name: "allowed to fail, with - prefix",
			in:   []string{"-/usr/bin/flup -a -T 32"},
			expected: []Exec{
				{
					Cmd:           "/usr/bin/flup",
					Args:          []string{"-a", "-T", "32"},
					AllowedToFail: true,
				},
			},
		},
		{
			name: "quoted args, single",
			in:   []string{"/bin/sh -c \"echo 'hello world'\""},
			expected: []Exec{
				{
					Cmd:  "/bin/sh",
					Args: []string{"-c", "echo 'hello world'"},
				},
			},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseCmd(subtest.in, subtest.inAllowEmpty)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if !reflect.DeepEqual(out, subtest.expected) {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestParseType(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		expected    Type
		expectedErr bool
	}{
		{
			name:     "simple type, lower case",
			in:       "Simple",
			expected: TypeSimple,
		},
		{
			name:     "oneshot type",
			in:       "oNeShOT",
			expected: TypeOneshot,
		},
		{
			name:        "unsupported type",
			in:          "guava",
			expectedErr: true,
		},
		{
			name:     "simple type, with newline",
			in:       "simple\n",
			expected: TypeSimple,
		},
		{
			name:     "empty, use simple",
			in:       "",
			expected: TypeSimple,
		},
		{
			name:     "notify type",
			in:       "notify",
			expected: TypeNotify,
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseType(subtest.in)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if out != subtest.expected {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestParseDur(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		inRequired  bool
		inDefault   time.Duration
		expected    time.Duration
		expectedErr error
	}{
		{
			name:     "non-empty",
			in:       "1337s",
			expected: 1337 * time.Second,
		},
		{
			name:        "required but empty",
			in:          "",
			inRequired:  true,
			expectedErr: ErrUnsupportedValue{},
		},
		{
			name:      "empty use default",
			in:        "",
			inDefault: 9000 * time.Second,
			expected:  9000 * time.Second,
		},
	}

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			if out, err := parseDur(st.in, "testoption=", st.inRequired, st.inDefault); err != nil {
				// note: TypeOf used instead of errors.Is since an instance of
				// the error is stored in the subtest (expectedErr), and I'm
				// unaware of any way to "store a type" for the comparison here
				if reflect.TypeOf(err) != reflect.TypeOf(st.expectedErr) {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
			} else {
				if st.expectedErr != nil {
					t.Fatalf("err expected: %T, err received: %T", st.expectedErr, err)
				}
				if out != st.expected {
					t.Fatalf("expected: %v, got: %v", st.expected, out)

				}
			}
		})
	}
}

func TestParseRestart(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		expected    RestartCond
		expectedErr bool
	}{
		{
			name:     "always, lower case",
			in:       "always",
			expected: RestartAlways,
		},
		{
			name:     "always, crazy case",
			in:       "AlWaYs",
			expected: RestartAlways,
		},
		{
			name:     "never",
			in:       "nO",
			expected: RestartNever,
		},
		{
			name:     "failure",
			in:       "on-failure",
			expected: RestartFailure,
		},
		{
			name:     "success",
			in:       "on-Success",
			expected: RestartSuccess,
		},
		{
			name:        "unsupported restart condition",
			in:          "dragon fruit",
			expectedErr: true,
		},
		{
			name:     "success, with newline",
			in:       "on-success\n",
			expected: RestartSuccess,
		},
		{
			name:     "empty, use never",
			in:       "",
			expected: RestartNever,
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseRestart(subtest.in)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if out != subtest.expected {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestParseDeps(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		expected    []string
		expectedErr bool
	}{
		{
			name:     "basic no tricks",
			in:       "foo bar bazz",
			expected: []string{"foo", "bar", "bazz"},
		},
		{
			name: "empty",
			in:   "",
		},
		{
			name:     "with .service ext",
			in:       "foo.service bar bazz.SERVICE",
			expected: []string{"foo", "bar", "bazz"},
		},
		{
			name:        "with unknown ext",
			in:          "foo bar.whatever bazz",
			expectedErr: true,
		},
		{
			name:     "basic, with newline",
			in:       "foo bar bazz\n",
			expected: []string{"foo", "bar", "bazz"},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseDeps(subtest.in)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if !reflect.DeepEqual(out, subtest.expected) {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestParseInt(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		expected    int
		defaultVal  int
		expectedErr bool
	}{
		{
			name:     "boring int",
			in:       "100",
			expected: 100,
		},
		{
			name:     "int w/ newline",
			in:       "1337\n\n",
			expected: 1337,
		},
		{
			name:       "empty val w/ default",
			in:         "\n",
			defaultVal: 4,
			expected:   4,
		},
		{
			name:        "not an int",
			in:          "one million dollars",
			expectedErr: true,
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseInt(subtest.in, subtest.defaultVal)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if out != subtest.expected {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestMultiline2Single(t *testing.T) {
	subtests := []struct {
		name     string
		in       string
		expected string
	}{
		{
			name:     "multiline with single backslash",
			in:       "this \\is a test\n",
			expected: "this is a test",
		},
		{
			name:     "single line",
			in:       "avocados are neat",
			expected: "avocados are neat",
		},
		{
			name:     "newline char",
			in:       "AllowMultilineValues enables\n a property",
			expected: "AllowMultilineValues enables a property",
		},
	}

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			out := multiline2single(st.in)
			if out != st.expected {
				t.Fatalf("expected: %v, got: %v", st.expected, out)
			}
		})
	}
}

func TestParseEnv(t *testing.T) {
	subtests := []struct {
		name        string
		in          string
		expected    map[string]string
		expectedErr bool
	}{
		{
			name: "basic no tricks",
			in:   "FOO=foo BAR=bar BAZZ=",
			expected: map[string]string{
				"FOO":  "foo",
				"BAR":  "bar",
				"BAZZ": "",
			},
		},
		{
			name: "quoted values",
			in:   "\"FOO=foo\" BAR=\"bar bazz\" BAZZ='figs quince' \"QUOTED=maybe so\"",
			expected: map[string]string{
				"FOO":    "foo",
				"BAR":    "bar bazz",
				"BAZZ":   "figs quince",
				"QUOTED": "maybe so",
			},
		},
		{
			name: "multiline",
			in:   "FOO=foo \\ \n BAR=\"bar bazz\"",
			expected: map[string]string{
				"FOO": "foo",
				"BAR": "bar bazz",
			},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			out, err := parseEnv(subtest.in)
			if err != nil {
				if !subtest.expectedErr {
					t.Fatalf("unexpected error: %s", err)
				} else {
					switch err.(type) {
					case ErrUnsupportedValue:
						// this was the test, congrats you passed
						return
					default:
						t.Fatalf("unexpected error: %s", err)
					}
				}
			} else if subtest.expectedErr {
				t.Fatalf("expected parsing error but did not receive one")
			}

			if !reflect.DeepEqual(out, subtest.expected) {
				t.Fatalf("expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}
