// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package env

import (
	"sync"
)

type EnvStore struct {
	sync.RWMutex
	store map[string]string
}

var envStore EnvStore
var once sync.Once

// get a handle to the environment store. This is a singleton.
func Init() *EnvStore {
	once.Do(func() {
		envStore.store = make(map[string]string)
	})

	return &envStore
}

// iterate all variable/value pairs in the environment store, sending the
// variable and value over the returned channel
func (e *EnvStore) IterVars() <-chan [2]string {
	ch := make(chan [2]string)
	go func() {
		e.RLock()
		defer e.RUnlock()

		for k, v := range e.store {
			ch <- [2]string{k, v}
		}
		close(ch)
	}()
	return ch
}

// get the value of the given environment variable, ok indicates whether it was
// found in the environment store
func (e *EnvStore) Get(v string) (val string, ok bool) {
	e.RLock()
	defer e.RUnlock()

	val, ok = e.store[v]
	return
}

// set the given environment variable to the given value
func (e *EnvStore) Set(v string, value string) {
	e.Lock()
	defer e.Unlock()

	e.store[v] = value
}
