// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package log

import (
	"fmt"
	"io"
	stdLog "log"
	"os"
	"sync"

	"github.com/hashicorp/logutils"
	"sr.ht/~craftyguy/superd/internal/color"
)

type Logger struct {
	log     *stdLog.Logger
	filter  *logutils.LevelFilter
	noColor bool
	sync.Mutex
}

const (
	DEBUG logutils.LogLevel = "DEBUG"
	ERROR logutils.LogLevel = "ERROR"
	INFO  logutils.LogLevel = "INFO"
	WARN  logutils.LogLevel = "WARN"
)

var logger *Logger

// Initialize a new logger writing the given io.Writer. This is a singleton, so
// if New() was called previously, the previously initialized object is
// returned.
func New(out io.Writer, noColor bool) *Logger {
	if logger == nil {
		logger = &Logger{
			filter: &logutils.LevelFilter{
				Levels:   []logutils.LogLevel{DEBUG, INFO, WARN, ERROR},
				MinLevel: logutils.LogLevel(INFO),
				Writer:   out,
			},
			noColor: noColor,
		}
		logger.log = stdLog.New(logger.filter, "", stdLog.Ldate|stdLog.Ltime)
	}
	return logger
}

// Initialize a new logger writing to stderr. This is a singleton, so if New()
// was called previously, then that object will be returned and not necessarily
// one initialized to write to stderr.
func Default() *Logger {
	return New(os.Stderr, false)
}

func Get() *Logger {
	return logger
}

// Sets the minimum logging level
func (l *Logger) SetMinLevel(level logutils.LogLevel) {
	l.Lock()
	defer l.Unlock()
	l.filter.SetMinLevel(level)
}

func (l *Logger) Printf(p string, f string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Print(format(p, INFO, l.noColor, fmt.Sprintf(f, s...)))
}

func (l *Logger) Println(p string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Println(format(p, INFO, l.noColor, s...))
}

func (l *Logger) Debugf(p string, f string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Print(format(p, DEBUG, l.noColor, fmt.Sprintf(f, s...)))
}

func (l *Logger) Debugln(p string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Println(format(p, DEBUG, l.noColor, s...))
}

func (l *Logger) Warnf(p string, f string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Print(format(p, WARN, l.noColor, fmt.Sprintf(f, s...)))
}

func (l *Logger) Warnln(p string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Println(format(p, WARN, l.noColor, s...))
}

func (l *Logger) Errorf(p string, f string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Print(format(p, ERROR, l.noColor, fmt.Sprintf(f, s...)))
}

func (l *Logger) Errorln(p string, s ...interface{}) {
	l.Lock()
	defer l.Unlock()
	l.log.Println(format(p, ERROR, l.noColor, s...))
}

func format(p string, l logutils.LogLevel, disableColor bool, s ...interface{}) string {
	label := fmt.Sprintf("[%s]", l)
	if !disableColor {
		switch l {
		case ERROR:
			label = color.Red(label)
		case WARN:
			label = color.Yellow(label)
		case DEBUG:
			label = color.Blue(label)
		case INFO:
			label = color.Cyan(label)
		}
	}
	return fmt.Sprintf("%s %s: ", p, label) + fmt.Sprint(s...)
}
