// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package manager

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"sort"
	"strings"
	"time"

	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/dag"
	"sr.ht/~craftyguy/superd/internal/env"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/internal/misc"
	"sr.ht/~craftyguy/superd/internal/service"
	"sr.ht/~craftyguy/superd/internal/xdg"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Manager struct {
	services       map[string]*service.Service
	depGraph       *dag.DAG
	config         map[string]*config.ServiceConfig
	extraConfigDir string

	// used to signal service goroutines to quit
	quit chan struct{}

	// signal reload of service config by sending an error chan
	reload chan chan error

	// whether services should autostart
	autostart bool

	logger *log.Logger
}

var prefix = "manager"

func Init(extraServiceDir string, autostart bool) *Manager {

	m := &Manager{
		config:         make(map[string]*config.ServiceConfig),
		services:       make(map[string]*service.Service),
		extraConfigDir: misc.ExpandHome(extraServiceDir),
		quit:           make(chan struct{}),
		reload:         make(chan chan error),
		autostart:      autostart,
		logger:         log.Get(),
	}

	return m
}

// returns a list of service config directories in order from lowest to highest
// priority
func (m *Manager) configDirs() (dirs []string) {
	// XDG_DATA_DIRS are the lowest priority
	dirs = append(dirs, xdg.DataDirs()...)

	// /etc is next lowest priority
	dirs = append(dirs, "/etc/superd/services")

	// $XDG_HOME is next lowest priority
	userConfDir, err := os.UserConfigDir()
	if err != nil {
		m.logger.Warnln(prefix, "unable to find user config directory, is $HOME unset?")
	}
	userConfEnabledDir := filepath.Join(userConfDir, "services", "enabled")

	if err := os.MkdirAll(userConfEnabledDir, 0750); err != nil {
		m.logger.Warnf(prefix, "unable to create directory %q: %s", userConfEnabledDir, err)
	} else {
		// include the parent dir, so configs are available for starting
		// manually, lower priority than the "enabled" subdir
		dirs = append(dirs, filepath.Join(userConfDir, "services"))
		dirs = append(dirs, userConfEnabledDir)
	}

	// highest priority directory
	if m.extraConfigDir != "" {
		dirs = append(dirs, m.extraConfigDir)
	}

	return
}

// interface that is meant to wrap filepath.Glob to make it easier to test methods that
// use this easier
type fileLister interface {
	listFiles(pattern string) ([]string, error)
}

type lister struct{}

func (l *lister) listFiles(pattern string) (files []string, err error) {
	return filepath.Glob(pattern)
}

// for mocking in testing
var osStat = os.Stat

// get configs from the given directories, sorted in order of priority from
// lowest to highest
func (m *Manager) collectConfig(searchDirs []string, lister fileLister) (configs []string) {
	// use a map to store the config file paths, allowing for easy 'overriding'
	// when a higher priority dir provides a file with the same name
	confMap := make(map[string]string)

	var masked []string

	// in order of preference/priority
	for _, dir := range searchDirs {
		path := filepath.Join(dir, "*.service")
		m.logger.Println(prefix, "searching dir: ", path)

		files, err := lister.listFiles(filepath.Join(dir, "*.service"))
		if err != nil {
			m.logger.Warnln(prefix, "unable to search dir: ", dir)
			continue
		}

		if len(files) == 0 {
			m.logger.Debugln(prefix, "no files found in dir: ", dir)
			continue
		}

		for _, file := range files {
			name := filepath.Base(strings.TrimSuffix(file, ".service"))
			// skip files that are templates
			if strings.HasSuffix(file, "@.service") {
				m.logger.Debugln(prefix, "skipping template file: ", file)
				continue
			}

			// skip empty config files (e.g. they are "masked")
			stat, err := osStat(file)
			if err != nil {
				m.logger.Warnf(prefix, "skipping %q because error getting stats: %s\n", file, err)
				continue
			}

			if stat.Size() == 0 {
				m.logger.Debugln(prefix, "found masked file: ", file)
				masked = append(masked, name)
				continue
			}

			confMap[name] = file
		}
	}

	configs = make([]string, 0, len(confMap))

confLoop:
	for name, file := range confMap {

		for _, maskedName := range masked {
			if maskedName == name {
				m.logger.Debugln(prefix, "masking service: ", name)
				continue confLoop
			}
		}

		configs = append(configs, file)
	}

	return
}

// Query status for the given list of service names, and return a map of
// service:api.ServiceStatus
// If services is an empty slice, it will return the status of all tracked services
func (m *Manager) Status(services []string) (statuses map[string]api.ServiceStatus, err error) {

	queryList := []*service.Service{}

	// print status of all services if none was specified
	if len(services) == 0 {
		for _, v := range m.services {
			queryList = append(queryList, v)
		}
		// or print status of the specified service
	} else {
		for _, name := range services {
			s, ok := m.services[name]
			if !ok {
				err = fmt.Errorf("service not found: %s", name)
				return
			}
			queryList = append(queryList, s)
		}
	}

	statuses = make(map[string]api.ServiceStatus)
	for _, s := range queryList {
		name := s.Name()

		var status api.ServiceStatus
		if status, err = s.Status(); err != nil {
			err = fmt.Errorf("manager.Status: %w", err)
			return
		}

		statuses[name] = status
	}

	return
}

// Start the given service
func (m *Manager) StartService(name string) (err error) {
	s, ok := m.services[name]
	if !ok {
		err = fmt.Errorf("manager.StartService: service not found: %s", name)
	} else {
		err = m.start(s, nil)
	}
	return
}

// Stop the given service
func (m *Manager) StopService(name string) (err error) {
	s, ok := m.services[name]
	if !ok {
		err = fmt.Errorf("manager.Stop: service not found: %s", name)
	} else {

		for _, dep := range s.Before() {
			err = m.StopService(dep)
			if err != nil {
				return
			}
		}
		m.logger.Printf(prefix, "stopping %q", name)
		select {
		case s.Stop <- true:
		case <-time.After(500 * time.Millisecond):
			// this *really* shouldn't time out unless there's something in the
			// service supervisor's main loop is blocking... which it
			// shouldn't...
			err = fmt.Errorf("manager.Stop: timed out trying to instruct service to stop: %s", name)
		}
	}
	return
}

// Restart the given service. If the service is stopped, this just starts it.
// This method can block for up to 10 seconds as it waits for the service to
// stop.
func (m *Manager) RestartService(name string) (err error) {
	status, err := m.Status([]string{name})
	if err != nil {
		return
	}

	if status[name].State == api.ServiceStarted {
		err = m.StopService(name)
		if err != nil {
			return
		}
	}
	s := m.services[name]
	timeout := time.After(10 * time.Second)
stop_loop:
	for {
		select {
		case <-timeout:
			err = fmt.Errorf("manager.Restart: timed out waiting for service to stop: %s", name)
			m.logger.Errorln(err.Error())
			return err
		default:
			var status api.ServiceStatus
			if status, err = s.Status(); err != nil {
				err = fmt.Errorf("manager.Restart: %w", err)
				m.logger.Errorln(prefix, err.Error())
				return
			}

			if status.State != api.ServiceStarted {
				break stop_loop
			}
			time.Sleep(100 * time.Millisecond)
		}
	}
	err = m.StartService(name)
	if err != nil {
		return
	}

	return
}

// Enable the given service name by creating a symlink from the service config
// in XDG_DATA_DIRS to XDG_CONFIG_HOME/services/enabled
func (m *Manager) EnableService(name string) (result string, err error) {
	cfgHome, err := os.UserConfigDir()
	if err != nil {
		err = fmt.Errorf("unable to get $XDG_CONFIG_HOME: %w", err)
		return
	}

	dataDirs := xdg.DataDirs()

	// XDG_CONFIG_HOME/services
	dataDirs = append(dataDirs, filepath.Join(cfgHome, "services"))

	// search for config using service name.
	// Note: services that use template files will require a .service file that
	// doesn't match the full name of the given service, e.g. "foo@bar", so
	// this needs to be handled here too by stripping off the characters after
	// @ when searching for the source .service file to use.
	serviceNameBase := name + ".service"
	if strings.Contains(name, "@") {
		serviceNameParts := strings.SplitAfterN(name, "@", 2)
		if serviceNameParts[1] == "" {
			err = fmt.Errorf("service uses a template, but no param was specified after '@'")
			return
		}
		serviceNameBase = serviceNameParts[0] + ".service"
	}

	source := ""
	for _, d := range dataDirs {
		path := filepath.Join(d, serviceNameBase)
		_, err := os.Stat(path)
		if err != nil {
			continue
		}
		// allows for later dirs in list to override path from previous dir(s)
		source = path
	}

	if source == "" {
		err = fmt.Errorf("unable to find config for service in: %s", name)
		return
	}

	dest := filepath.Join(cfgHome, "services", "enabled", name+".service")

	if err = os.MkdirAll(filepath.Dir(dest), 0750); err != nil {
		err = fmt.Errorf("unable create dir %q: %w", filepath.Base(dest), err)
		return
	}

	// clean up any old symlinks for service in the enabled dir
	if err := os.RemoveAll(filepath.Join(cfgHome, "services", "enabled", name+".service")); err != nil {
		err = fmt.Errorf("unable to remove old service file %w", err)
		return "", err
	}

	if err = os.Symlink(source, dest); err != nil {
		err = fmt.Errorf("unable create symlink %s --> %s: %w", source, dest, err)
		return
	}

	// reload config to update any status
	if err = m.Reload(); err != nil {
		return
	}

	result = fmt.Sprintf("create symlink %s --> %s", source, dest)

	return
}

// Disable auto-starting the given service
func (m *Manager) DisableService(name string) (result string, err error) {

	cfgHome, err := os.UserConfigDir()
	if err != nil {
		err = fmt.Errorf("unable to get $XDG_CONFIG_HOME: %w", err)
		return
	}

	// if service config exists in /etc, mask it by creating a symlink
	// /dev/null-->XDG_CONFIG_HOME/services/enabled
	if _, err := os.Stat(filepath.Join("/etc/superd/services", name+".service")); err == nil {
		// source for "masking" the service, using a symlink to this location
		source := "/dev/null"

		dest := filepath.Join(cfgHome, "services", "enabled", name+".service")

		if err = os.MkdirAll(filepath.Base(dest), 0750); err != nil {
			err = fmt.Errorf("unable create dir %q: %w", filepath.Base(dest), err)
			return "", err
		}

		// clean up any old symlinks for service in the enabled dir
		if err := os.RemoveAll(filepath.Join(cfgHome, "services", "enabled", name+".service")); err != nil {
			err = fmt.Errorf("unable to remove old service file %w", err)
			return "", err
		}

		if err = os.Symlink(source, dest); err != nil {
			err = fmt.Errorf("unable create symlink %s --> %s: %w", source, dest, err)
			return "", err
		}

		result = fmt.Sprintf("masked service from auto-starting: %s", name)

	} else if _, err := os.Stat(filepath.Join(cfgHome, "services", "enabled", name+".service")); err == nil {
		// remove symlink in "enabled"
		if err := os.RemoveAll(filepath.Join(cfgHome, "services", "enabled", name+".service")); err != nil {
			err = fmt.Errorf("unable to remove old service file %w", err)
			return "", err
		}
		result = fmt.Sprintf("disabled service from auto-starting: %s", name)
	}

	// reload config to update any status
	if err = m.Reload(); err != nil {
		return
	}

	return
}

func (m *Manager) Reload() error {
	errCh := make(chan error)
	defer close(errCh)

	m.reload <- errCh

	if err := <-errCh; err != nil {
		return err
	}

	return nil
}

func (m *Manager) loadServiceConfig(s *service.Service, c *config.ServiceConfig) error {
	s.SetConfig(*c)

	return nil
}

// Scans for new or updated service config, and starts supervising goroutines
// for any new config so that the services for them can be managed by this
// current session.
func (m *Manager) load() error {

	dirs := m.configDirs()
	configs := m.collectConfig(dirs, &lister{})

	userConfDir, err := os.UserConfigDir()
	if err != nil {
		m.logger.Warnln(prefix, "unable to find user config directory, is $HOME unset?")
	}
	// services in these dirs are not auto-started, but the config is still
	// loaded so they can be started manually
	noAutoDirs := xdg.DataDirs()
	noAutoDirs = append(noAutoDirs, filepath.Join(userConfDir, "services"))

	// add configs for each service found
	m.logger.Println(prefix, "Using files:")
	sort.Strings(configs)
	for _, file := range configs {
		c, err := config.Parse(file)
		if err != nil {
			return err
		}

		// New service
		if _, ok := m.services[c.Name]; !ok {
			m.services[c.Name] = service.Init(c.Name)
			go m.services[c.Name].Run(m.quit)
		}
		c.Autostart = true
		// disable auto-start if service is in a dir where services should not
		// be auto-started
		for _, d := range noAutoDirs {
			if d == filepath.Dir(file) {
				c.Autostart = false
			}
		}

		m.config[c.Name] = c

		// Load config in services
		if err := m.loadServiceConfig(m.services[c.Name], c); err != nil {
			m.logger.Errorln(prefix, err.Error())
			// TODO: really return here? or just log error and continue?
			return err
		}

		m.logger.Println(prefix, "--", file)
	}

	// Compute dependencies
	err = m.calcServiceDepends()
	if err != nil {
		m.logger.Errorln(prefix, err.Error())
		return err
	}

	return nil
}

// Set environment variables to be used when starting/restarting services in
// the future. Note: has no impact on the environment for already-running
// services.
func (m *Manager) SetEnv(e map[string]string) {
	envStore := env.Init()
	for k, v := range e {
		envStore.Set(k, v)
	}
}

// Starts the given service, after starting all dependencies (if they are not
// already started).
func (m *Manager) start(s *service.Service, cancel <-chan struct{}) (err error) {
	name := s.Name()

	// TODO: make specific timeout error, failed error. maybe in services?
	for _, dep := range s.After() {
		if _, ok := m.services[dep]; !ok {
			err = fmt.Errorf("%s: unable to start, missing dependency: %s", name, dep)
			m.logger.Errorf(name, err.Error())
			return
		}
		m.logger.Printf(prefix, "%s: depends on %q, starting dependency", name, dep)
		go m.start(m.services[dep], cancel)
	wait_dep_loop:
		for {
			select {
			case <-time.After(100 * time.Millisecond):
				var status api.ServiceStatus
				if status, err = m.services[dep].Status(); err != nil {
					err = fmt.Errorf("failed to get status: %w", err)
					m.logger.Errorf(name, err.Error())
					return
				}
				switch status.State {
				case api.ServiceStarted:
					break wait_dep_loop
				case api.ServiceStopped:
					err = fmt.Errorf("unable to start, dependency in stopped state: %s", dep)
					m.logger.Errorf(name, err.Error())
					return
				case api.ServiceFailed:
					err = fmt.Errorf("unable to start, dependency in failed state: %s", dep)
					m.logger.Errorf(name, err.Error())
					return
				}
			case <-time.After(10 * time.Second):
				err = fmt.Errorf("unable to start, timed out waiting for dependency to start: %s", dep)
				m.logger.Errorf(name, err.Error())
				return
			case <-cancel:
				return
			}
		}
	}

	m.logger.Printf(prefix, "starting %q", name)
	s.Start <- true
	for {
		select {
		case <-time.After(100 * time.Millisecond):
			var status api.ServiceStatus
			if status, err = s.Status(); err != nil {
				err = fmt.Errorf("failed to get status: %w", err)
				m.logger.Errorf(name, err.Error())
				return
			}
			switch status.State {
			case api.ServiceFailed:
				err = fmt.Errorf("failed to start")
				m.logger.Errorf(name, err.Error())
				return
			case api.ServiceStarted:
				return
			}
		case <-time.After(5 * time.Second):
			err = fmt.Errorf("timed out waiting to start")
			m.logger.Errorf(name, err.Error())
			return
		case <-cancel:
			return
		}
	}

}

func (m *Manager) buildDepGraph() (err error) {

	m.depGraph = dag.NewDag()

	// Add to dep. graph. This loop is separate from the one that
	// discovers service config since config might be overwritten by conf dir
	// heirarchy
	for u := range m.services {
		// TODO: add support for "requires" and "wants"
		for _, a := range m.config[u].After {
			if _, ok := m.services[a]; !ok {
				err = fmt.Errorf("error: unable to find missing dependency for %q: %s", u, a)
				return err
			}
			if err := m.depGraph.AddEdge(a, u); err != nil {
				if errors.Is(err, dag.ErrDAGCycle) {
					err = fmt.Errorf("%w detected for %s, %s", err, a, u)
				} else {
					err = fmt.Errorf("unknown error: %w", err)
				}
				return err
			}
		}

		for _, b := range m.config[u].Before {
			if _, ok := m.services[b]; !ok {
				// No need to fail here if the service has to start *before*
				// something else. that thing might not exist, and that's fine.
				// If it does exist, it'll test for the existence of this
				// service
				continue
			}

			if err := m.depGraph.AddEdge(u, b); err != nil {
				if errors.Is(err, dag.ErrDAGCycle) {
					err = fmt.Errorf("%w detected for %s, %s", err, u, b)
				} else {
					err = fmt.Errorf("unknown error: %w", err)
				}
				return err
			}
		}
	}

	return
}

func (m *Manager) calcServiceDepends() (err error) {

	if err = m.buildDepGraph(); err != nil {
		return
	}

	// resolve all dependencies for services here, since all services should be
	// added at this point
	for _, s := range m.services {
		for _, b := range m.depGraph.AdjacentNodes(s.Name()) {
			// services in the dag that point from this service are services that
			// depend on this service. This service should be started *before*
			// those.
			s.AddBefore(m.services[b].Name())

			// services that this service directly depends on should be
			// started *after* this service
			m.services[b].AddAfter(s.Name())
		}
	}
	return
}

// Starts the manager, and quits when the given done channel is closed. If
// given a ready channel that is not nil, it will close the ready channel after
// config has been loaded and the main loop of the manager has been entered,
// indicating that the manager is ready to accept commands via method calls.
func (m *Manager) Start(done <-chan struct{}, ready chan<- struct{}) (err error) {

	err = m.load()
	if err != nil {
		return err
	}

	// set some default environment variables
	e := env.Init()

	if home, err := os.UserHomeDir(); err == nil {
		e.Set("HOME", home)
	}
	vars := []string{"XDG_RUNTIME_DIR", "USER", "SHELL", "PATH"}
	for _, v := range vars {
		val, ok := os.LookupEnv(v)
		if ok {
			e.Set(v, val)
		}
	}

	// start all services
	if m.autostart {
		for _, v := range m.services {
			if v.Autostart() {
				go m.start(v, done)
			}
		}
	} else {
		m.logger.Println(prefix, "autostarting of services has been DISABLED")
	}

	// TODO: make configurable?
	timeoutDur := time.Duration(10 * time.Second)

	for {
		// signal that the manager is open for business
		if ready != nil {
			close(ready)
			ready = nil
		}

		select {
		case errCh := <-m.reload:
			errCh <- m.load()
		case <-done:
			m.logger.Println(prefix, "quit requested")
			// tell all services to stop
			for _, s := range m.services {
				err := m.StopService(s.Name())
				if err != nil {
					m.logger.Errorf(prefix, "error received when trying to stop service: %s\n", err.Error())
				}
			}
			timeout := time.After(timeoutDur)
		quit_loop:
			// wait for all services to be stopped, with a hard timeout if some of
			// them fail to reach that point
			for {
				select {
				case <-timeout:
					m.logger.Printf(prefix, "some processes failed to stop after %s... quitting anyways!\n", timeoutDur.String())
					break quit_loop
				default:
					stopped := true
					// make sure no services are "started" (while allowing for
					// them to be in other states, like "failed")
					for _, s := range m.services {
						var status api.ServiceStatus
						if status, err = s.Status(); err != nil {
							m.logger.Errorf(prefix, "%s", err.Error())
						}
						stopped = stopped && status.State != api.ServiceStarted
					}
					if stopped {
						break quit_loop
					}
					time.Sleep(100 * time.Millisecond)
				}
			}
			close(m.quit)
			return
		}
	}
}
