// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package manager

import (
	"errors"
	"io/fs"
	"os"
	"path/filepath"
	"reflect"
	"sort"
	"testing"

	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/dag"
	"sr.ht/~craftyguy/superd/internal/log"
)

// mock fileLister
type mockLister struct {
	files map[string]int64
}

func (m *mockLister) listFiles(pattern string) (files []string, err error) {
	list := []string{}
	dir := filepath.Dir(pattern)
	// only return files in the dir that would match the given pattern's dir
	for f := range m.files {
		if filepath.Dir(f) == dir {
			list = append(list, f)
		}
	}
	return list, nil
}

// mock fs.Info
type mockInfo struct {
	fs.FileInfo
	size int64
}

func (i *mockInfo) Size() int64 {
	return i.size
}

func TestCollectConfig(t *testing.T) {
	subtests := []struct {
		name string
		// in order of priority, lowest to highest
		in_dirs []string
		// .service files in dirs, int64 == size (0 is 'masked')
		in_files map[string]int64
		// final ordering of config, accounting for priority
		expected []string
	}{
		{
			name: "multiple priorities w/ templates",
			// in order of priority, lowest to highest
			in_dirs: []string{
				"/etc/superd/services",
				"/home/clayton/.config/services",
				"/home/clayton/.config/services/enabled",
				"/tmp/bazz",
			},
			in_files: map[string]int64{
				"/etc/superd/services/foo.service":                            100,
				"/etc/superd/services/bar.service":                            100,
				"/etc/superd/services/bar@.service":                           900,
				"/etc/superd/services/foo@bar.service":                        10000,
				"/home/clayton/.config/services/enabled/foo.service":          100,
				"/home/clayton/.config/services/enabled/goimapnotify.service": 100,
				"/tmp/bazz/foo.service":                                       100,
				"/etc/superd/services/goimapnotify.service":                   100,
				"/etc/superd/services/kanshi.service":                         100,
				"/home/clayton/.config/services/enabled/thing@me.service":     100,
				"/home/clayton/.config/services/thing@me.service":             100,
				"/etc/superd/services/thing@.service":                         100,
				"/home/clayton/.config/services/mango.service":                100,
				// test masking
				"/etc/superd/services/shouldbemasked.service":                   100,
				"/home/clayton/.config/services/enabled/shouldbemasked.service": 0,
			},
			expected: []string{
				"/etc/superd/services/bar.service",
				"/etc/superd/services/foo@bar.service",
				"/home/clayton/.config/services/enabled/goimapnotify.service",
				"/tmp/bazz/foo.service",
				"/etc/superd/services/kanshi.service",
				"/home/clayton/.config/services/enabled/thing@me.service",
				"/home/clayton/.config/services/mango.service",
			},
		},
	}

	for _, subtest := range subtests {

		osStat = func(name string) (fs.FileInfo, error) {
			return &mockInfo{size: subtest.in_files[name]}, nil
		}
		defer func() { osStat = os.Stat }()

		t.Run(subtest.name, func(t *testing.T) {
			m := Init("", false)
			// logger will not have been initialized yet, so set it
			m.logger = log.Default()

			out := m.collectConfig(subtest.in_dirs, &mockLister{files: subtest.in_files})

			sort.Strings(out)
			sort.Strings(subtest.expected)

			if !reflect.DeepEqual(out, subtest.expected) {
				t.Fatalf("Expected: %v, got: %v", subtest.expected, out)
			}
		})
	}
}

func TestBuildDepGraph(t *testing.T) {
	subtests := []struct {
		name         string
		in           map[string]*config.ServiceConfig
		expected     map[string][]string
		expected_err error
	}{
		{
			name: "3 services",
			in: map[string]*config.ServiceConfig{
				"service1": {
					Before: []string{"service2", "service3"},
				},
				"service2": {
					Before: []string{"service3"},
				},
				"service3": {
					After: []string{"service1"},
				},
			},
			expected: map[string][]string{
				"service1": {"service2", "service3"},
				"service2": {"service3"},
			},
		},
		{
			name: "7 services",
			in: map[string]*config.ServiceConfig{
				"service1": {
					Before: []string{"service2", "service3"},
				},
				"service2": {
					Before: []string{"service3"},
				},
				"service3": {
					After: []string{"service1"},
				},
			},
			expected: map[string][]string{
				"service1": {"service2", "service3"},
				"service2": {"service3"},
			},
		},
		{
			name: "circular dependency with Before",
			in: map[string]*config.ServiceConfig{
				"service1": {
					Before: []string{"service2", "service3"},
				},
				"service2": {
					Before: []string{"service3"},
				},
				"service3": {
					Before: []string{"service1"},
				},
			},
			expected_err: dag.ErrDAGCycle,
		},
		{
			name: "circular dependency with After",
			in: map[string]*config.ServiceConfig{
				"service1": {
					Before: []string{"service2", "service3"},
					After:  []string{"service3"},
				},
				"service2": {
					Before: []string{"service3"},
				},
				"service3": {},
			},
			expected_err: dag.ErrDAGCycle,
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			m := Init("", false)
			// logger will not have been initialized yet, so set it
			m.logger = log.Default()

			m.config = subtest.in
			for s := range m.config {
				// doesn't matter if nil, it's not accessed for this test
				m.services[s] = nil
			}

			if err := m.buildDepGraph(); err != nil {
				if !errors.Is(err, subtest.expected_err) {
					t.Fatalf("unexpected error: %s", err)
				}
				// this was the test, congrats you passed
				return
			}

			for s := range m.services {
				out := m.depGraph.AdjacentNodes(s)
				sort.Strings(out)
				sort.Strings(subtest.expected[s])
				if !reflect.DeepEqual(subtest.expected[s], out) {
					t.Fatalf("%s expected %v, got: %v", s, subtest.expected[s], out)
				}
			}
		})
	}
}
