// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"errors"
	"fmt"
	"os"
	"path/filepath"
	"strings"
	"sync"
	"time"

	goCmd "git.sr.ht/~craftyguy/go-cmd"
	"sr.ht/~craftyguy/superd/internal/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/env"
	"sr.ht/~craftyguy/superd/internal/log"
	"sr.ht/~craftyguy/superd/internal/misc"
	"sr.ht/~craftyguy/superd/pkg/api"
)

type Service struct {
	Start chan bool
	Stop  chan bool

	getStatus chan chan api.ServiceStatus

	reload chan bool
	logger *log.Logger
	// make cmd.NewCommand overridable in testing
	newCommand func(exe string, args []string, out *os.File) cmd.Commander

	before struct {
		list []string
		sync.RWMutex
	}
	after struct {
		list []string
		sync.RWMutex
	}

	name struct {
		value string
		sync.RWMutex
	}

	config struct {
		value config.ServiceConfig
		sync.RWMutex
	}
}

// whether or not the service should be autostarted
func (s *Service) Autostart() bool {
	return s.Config().Autostart
}

func (s *Service) Name() string {
	s.name.RLock()
	defer s.name.RUnlock()
	return s.name.value
}

func (s *Service) AddAfter(name string) {
	s.after.Lock()
	defer s.after.Unlock()
	for _, n := range s.after.list {
		if n == name {
			return
		}
	}
	s.after.list = append(s.after.list, name)
}

func (s *Service) AddBefore(name string) {
	s.before.Lock()
	defer s.before.Unlock()
	for _, n := range s.before.list {
		if n == name {
			return
		}
	}
	s.before.list = append(s.before.list, name)
}

func (s *Service) After() []string {
	s.after.RLock()
	defer s.after.RUnlock()
	// return a copy
	return append([]string{}, s.after.list...)
}

func (s *Service) Before() []string {
	s.before.RLock()
	defer s.before.RUnlock()
	// return a copy
	return append([]string{}, s.before.list...)
}

func (s *Service) SetConfig(c config.ServiceConfig) {
	s.config.Lock()
	defer s.config.Unlock()
	s.config.value = c
	go func() {
		select {
		case <-time.After(1 * time.Second):
			s.logger.Warnln(s.Name(), "timed out trying to signal config reload, this shouldn't have happened....")
		case s.reload <- true:
		}
	}()
}

func (s *Service) Config() config.ServiceConfig {
	s.config.RLock()
	defer s.config.RUnlock()
	return s.config.value
}

func (s *Service) Status() (status api.ServiceStatus, err error) {

	// TODO: this isn't thread-safe! maybe receive from it then timeout with a
	// specific (custom) timeout error?
	if s.getStatus != nil {
		response := make(chan api.ServiceStatus)
		select {
		case s.getStatus <- response:
			break
		case <-time.After(1 * time.Second):
			err = fmt.Errorf("timed out requesting status for: %s", s.Name())
			return
		}

		select {
		case d := <-response:
			status = d
		case <-time.After(1 * time.Second):
			err = fmt.Errorf("timed out getting status for: %s", s.Name())
			return
		}

	} else {
		// assume it is stopped if the channel was nil.
		// also note that if the service hasn't been Init'd yet, then no
		// config was loaded for it either.
		status = api.ServiceStatus{
			State: api.ServiceStopped,
		}
	}
	return
}

func Init(name string) (s *Service) {

	s = &Service{
		Start:     make(chan bool),
		Stop:      make(chan bool),
		getStatus: make(chan chan api.ServiceStatus),
		reload:    make(chan bool),
		logger:    log.Get(),
	}
	s.name.Lock()
	defer s.name.Unlock()
	s.name.value = name
	// make cmd.NewCommand overridable in testing
	s.newCommand = cmd.NewCommand

	return
}

// Runs the given command, if notify is true then it creates a socket for the
// service to notify when it has started.
func (s *Service) runCmd(c cmd.Commander, notify bool) (err error) {

	// environment vars are collected from various sources in order of
	// importance, from least to most, allowing later sources (e.g. service
	// config file) to override previous var values
	eMap := make(map[string]string)

	// add "global" environment
	for v := range env.Init().IterVars() {
		eMap[v[0]] = v[1]
	}

	// import all variables from current env
	for _, v := range os.Environ() {
		name, value, _ := strings.Cut(v, "=")
		eMap[name] = value
	}

	// set service-specific env. from config
	for k, v := range s.Config().Env {
		eMap[k] = v
	}

	// format env into a list
	var e []string
	for k, v := range eMap {
		e = append(e, fmt.Sprintf("%s=%s", k, v))
	}

	if notify {
		err = c.StartNotify(e, s.Name()+"_notify.sock", s.Config().TimeoutStartDur)
	} else {
		err = c.Start(e)
	}

	return
}

// startup service routine, e.g. running ExecStartPre, ExecStart. Returns
// channels for receiving the command object if successful or service status on
// failure. the cancel chan is used to signal to the routine to abort startup (e.g. when retrying)
func (s *Service) startup(cancel <-chan struct{}, logFile *os.File) (cmdChan chan cmd.Commander, startState chan api.ServiceState) {
	startState = make(chan api.ServiceState)
	cmdChan = make(chan cmd.Commander)

	go func() {
		defer close(startState)
		defer close(cmdChan)

		conf := s.Config()

		// ExecStartPre
		for _, exec := range conf.ExecStartPre {
			preCmd := s.newCommand(exec.Cmd, exec.Args, logFile)
			if err := s.runCmd(preCmd, false); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
				if exec.AllowedToFail {
					continue
				}
				startState <- api.ServiceFailed
				return
			}
		}

		var startCmd cmd.Commander
		// ExecStart

		// burst limit timing
		limitStart := time.Now()
		tries := 0

	start_loop:
		for _, exe := range conf.ExecStart {
			for {
				// if retry/burst limit is exceeded during the start limit duration, then fail
				if tries >= conf.StartLimitBurst {
					if time.Since(limitStart) < conf.StartLimitDur {
						s.logger.Errorf(s.Name(), "failed to start service: exceeded the max number of tries, %d, in %s",
							conf.StartLimitBurst, conf.StartLimitDur.String())
						startState <- api.ServiceFailed
						return
					}

					// burst limit wasn't exceeded over the burst duration, so reset and keep going
					limitStart = time.Now()
					tries = 0
				}

				startCmd = s.newCommand(exe.Cmd, exe.Args, logFile)
				if err := s.runCmd(startCmd, conf.Type == config.TypeNotify); err != nil {
					switch err.(type) {
					case cmd.ErrCmdExitedNonZero:
					case cmd.ErrUnhandledStatus:
					case cmd.ErrCmdExited:
						// oneshot is expected to exit
						if conf.Type == config.TypeOneshot {
							break start_loop
						}
					case cmd.ErrTimeout:
					default:
						s.logger.Errorln(s.Name(), "unknown failure when trying to start service: "+err.Error())
					}
					s.logger.Errorln(s.Name(), "failed to start service: "+err.Error())
				} else {
					break
				}

				// either the restart timer expires, or a signal is received to
				// cancel the startup
				select {
				case <-time.After(conf.RestartDur):
					tries++
					s.logger.Errorln(s.Name(), "retrying...")
					continue
				case <-cancel:
					s.logger.Debugln(s.Name(), "service startup canceled")
					return
				}
			} // run cmd retry loop
		} // ExecStart loop

		// ExecStartPost
		for _, exec := range conf.ExecStartPost {
			postCmd := s.newCommand(exec.Cmd, exec.Args, logFile)
			if err := s.runCmd(postCmd, false); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
				if exec.AllowedToFail {
					continue
				}
				startState <- api.ServiceFailed
				return
			}
		}

		// send the cmd from ExecStart
		cmdChan <- startCmd
	}()

	return
}

// make killTimeout overridable in testing
var killTimeout = time.Duration(5 * time.Second)

func (s *Service) newLogFile() (file *os.File, err error) {
	logDir, ok := os.LookupEnv("XDG_STATE_HOME")
	if !ok {
		logDir = "~/.local/state"
	}
	logDir = filepath.Join(misc.ExpandHome(logDir), "superd", "logs")
	logFilePath := filepath.Join(logDir, s.Name()+".log")

	if err = os.MkdirAll(logDir, 0750); err != nil {
		err = fmt.Errorf("unable to create dir %q for service log file: %w", logDir, err)
		return
	}

	// rotate any old log file to <name>.old
	// Note: Error is ignored since this may fail if the log doesn't exist. For
	// all other types of failures, we'll fail later when creating or writing
	// to it...
	_ = os.Rename(logFilePath, logFilePath+".old")

	file, err = os.Create(logFilePath)
	if err != nil {
		err = fmt.Errorf("unable to create service log file %q: %w", logFilePath, err)
		return
	}

	return
}

func (s *Service) Run(quit <-chan struct{}) {

	var (
		startState     <-chan api.ServiceState
		onFailComplete chan struct{}
		restartTimer   <-chan time.Time
		status         api.ServiceStatus
		// used for signaling that the service should *not* be restarted, e.g.
		// it was explicitly stopped
		shouldRestart bool
		// used for signaling that the service was explicitly stopped, so that
		// status can be reflected accurately
		shouldStop bool

		// used for receiving the started service process status
		cmdChan         <-chan cmd.Commander
		command         cmd.Commander
		cancelStartChan chan struct{}

		// command streams
		cmdStatusCh <-chan goCmd.Status

		//  for stdout/stderr
		logFile *os.File
	)
	status.State = api.ServiceStopped

	for {
		select {

		// Start (process) requested
		case <-s.Start:
			if status.State != api.ServiceStarted && status.State != api.ServiceStarting {

				// declare err, since using := results in a new logFile being
				// instantiated in this scope
				var err error
				logFile, err = s.newLogFile()
				if err != nil {
					s.logger.Errorln(err.Error())
				}

				shouldRestart = true
				shouldStop = false
				cancelStartChan = make(chan struct{})
				cmdChan, startState = s.startup(cancelStartChan, logFile)
				status.State = api.ServiceStarting
			}

		// new service process info received (e.g. from startup())
		case command = <-cmdChan:
			if command == nil {
				// startup was canceled
				status.State = api.ServiceStopped
				continue
			} else {
				close(cancelStartChan)
				cancelStartChan = nil
			}
			startState = nil
			cmdChan = nil
			streams := command.Streams()
			cmdStatusCh = streams.Status
			status.State = api.ServiceStarted
			status.Pid = command.Status().PID

		// service startup state
		case ss := <-startState:
			startState = nil
			cmdStatusCh = nil
			cmdChan = nil
			switch ss {
			case api.ServiceFailed:
				status.State = api.ServiceFailed
				// Run execOnFailure
				onFailComplete = make(chan struct{})
				s.execOnFail(onFailComplete, logFile)
				// Note: retries have already happened, so at this point the
				// service is "failed"

				// don't allow the service to be restarted, it already exceeded the
				// number of startup retries
				shouldRestart = false
			}

		// Stop (process) requested
		case <-s.Stop:
			// Cancel restart and start timer in so that it doesn't trigger and
			// restart the service
			restartTimer = nil

			// don't allow the service to be restarted
			shouldRestart = false

			// indicate that the service is stopped because it was told to
			shouldStop = true

			if cancelStartChan != nil {
				close(cancelStartChan)
				cancelStartChan = nil
			}

			if onFailComplete != nil {
				close(onFailComplete)
				onFailComplete = nil
			}

			// guard against many rapid calls to stop
			if command != nil {
				if s.Config().Type == config.TypeOneshot {
					// oneshot services can't be "stopped", there should be no
					// process runninng... so just indicate that the state is
					// "stopped" when requested to
					status.State = api.ServiceStopped
				} else if err := command.Stop(killTimeout); err != nil {
					s.logger.Errorln(s.Name(), "failed to stop: "+err.Error())
					status.State = api.ServiceFailed
				}
				// for other situations, the service's state is set when the
				// updated cmdStatusCh is received later
				command = nil
			}

		// Quit (goroutine) requested
		case <-quit:
			// Note: no cleanup happens here, so it's expected that the manager
			// will have signaled "Stop" before this channel is closed
			return

			// Send status/status
		case response := <-s.getStatus:
			if command != nil {
				status.Uptime = int64(command.Status().Runtime)
			}

			// don't block on sending response
			go func(status api.ServiceStatus) {
				select {
				case <-time.After(2 * time.Second):
					// note: timeout arbitrarily chosen..
					s.logger.Warnln("timed out trying to send a status response")
				case response <- status:
				}
			}(status)

		// Status from running process
		case cmdStatus := <-cmdStatusCh:

			// stopped
			if cmdStatus.StopTs > 0 {
				s.logger.Debugln(s.Name(), "exited with: "+fmt.Sprint(cmdStatus.Exit))

				if logFile != nil {
					if err := logFile.Close(); err != nil {
						s.logger.Errorf("unable to close service log file: %s", err.Error())
					}
				}

				failed := false
				if cmdStatus.Error != nil {
					failed = true
				} else if cmdStatus.Exit != 0 {
					// err from os/exec
					failed = true
				} else if !cmdStatus.Complete {
					// command didn't run to completion, it was stopped(?)
					// note: setting failed not strictly necessary here, but
					// it's readable.
					failed = false
				} else {
					// command completed successfully
					if s.Config().Type == config.TypeOneshot {
						// show 'started' for oneshot services
						status.State = api.ServiceStarted
					} else {
						status.State = api.ServiceStopped
					}
				}

				if failed {
					if shouldStop {
						// explicitly stopped services shouldn't have a 'failed'
						// state. This can happen when an app receives a SIGTERM
						// but doesn't handle it, and exits with something like -1.
						// Technically the app "failed" to run to completion, but
						// for our purposes it did stop successfully (i.e. it's not
						// running anymore).
						status.State = api.ServiceStopped
						continue
					}

					status.State = api.ServiceFailed

					// fire off any execOnFailure
					onFailComplete = make(chan struct{})
					s.execOnFail(onFailComplete, logFile)

					// don't schedule a restart until after the
					// ExecOnFailure command has completed
					continue
				} else {
					status.State = api.ServiceStopped
				}

				if shouldRestart {
					restartTimer = s.scheduleRestart(status.State)
				}

			} else {
				s.logger.Warnln(s.Name(), "received unhandled cmdStatus: "+fmt.Sprint(cmdStatus))
			}

		// onFailStatus -- from ExecOnFailure
		case <-onFailComplete:
			onFailComplete = nil
			// Restart scheduled if necessary
			if shouldRestart {
				restartTimer = s.scheduleRestart(status.State)
			}

		// Restart timer
		case <-restartTimer:
			go func() {
				select {
				case <-time.After(1 * time.Second):
					s.logger.Warnln(s.Name(), "timed out trying to start the service, this shouldn't have happened....")
				case s.Start <- true:
				}
			}()

		// config should be reloaded
		case <-s.reload:
			// update status. note that this is only updating fields that
			// aren't changed by starting/stopping the service, or other
			// runtime events related to the supervised service.
			status.Autostart = s.Autostart()
			status.ConfigPath = s.Config().Path
			status.Type = string(s.Config().Type)
			status.Description = s.Config().Description
		}
	}
}

func (s *Service) oneshotShouldRestart(state api.ServiceState) bool {
	// oneshot should only restart if it is set to Always or OnFailure, and
	// only when the state is "failed"
	if state == api.ServiceFailed {
		if s.Config().Restart == config.RestartAlways || s.Config().Restart == config.RestartFailure {
			return true
		}
	}
	return false
}

func (s *Service) simpleShouldRestart(state api.ServiceState) bool {
	switch state {
	case api.ServiceFailed:
		if s.Config().Restart == config.RestartAlways || s.Config().Restart == config.RestartFailure {
			return true
		}
	case api.ServiceStopped:
		if s.Config().Restart == config.RestartAlways {
			return true
		}
	}

	return false
}

// This exists purely to allow overriding in testing
var getRestartTimer = func(dur time.Duration) <-chan time.Time {
	return time.After(dur)
}

// Checks whether a restart should be scheduled based on the given state and
// the service's config, and returns a channel to a timer for when it should be
// restarted. The "timer" returned will have been started.
func (s *Service) scheduleRestart(state api.ServiceState) (timer <-chan time.Time) {
	var restart bool

	switch s.Config().Type {
	case config.TypeOneshot:
		restart = s.oneshotShouldRestart(state)
	case config.TypeSimple:
		restart = s.simpleShouldRestart(state)
	default:
		s.logger.Warnf(s.Name(), "unable to determine restart behavior, unknown service type: %s", s.Config().Type)
	}

	if restart {
		s.logger.Debugln(s.Name(), fmt.Sprintf("restarting in %d second(s)", uint64(s.Config().RestartDur.Seconds())))
		timer = getRestartTimer(s.Config().RestartDur)
		return
	}
	return
}

// Run the ExecOnFailure cmd and close the given channel when done.
// If no ExecOnFailure commands are specified, it immediately closes the given
// channel.
func (s *Service) execOnFail(done chan struct{}, logFile *os.File) {

	// ExecStartPre
	for _, exec := range s.Config().ExecOnFailure {
		c := s.newCommand(exec.Cmd, exec.Args, logFile)
		if err := s.runCmd(c, false); err != nil && !errors.As(err, &cmd.ErrCmdExited{}) {
			s.logger.Errorf(s.Name(), "failed to run ExecOnFailure %q: %s\n", c, err)
			return
		}
	}

	close(done)
}
