// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package service

import (
	"fmt"
	"os"
	"reflect"
	"sync"
	"sync/atomic"
	"testing"
	"time"

	goCmd "git.sr.ht/~craftyguy/go-cmd"
	"sr.ht/~craftyguy/superd/internal/cmd"
	"sr.ht/~craftyguy/superd/internal/config"
	"sr.ht/~craftyguy/superd/internal/log"
	superdTesting "sr.ht/~craftyguy/superd/internal/testing"
	"sr.ht/~craftyguy/superd/pkg/api"
)

func helperService(name string) *Service {
	s := Init(name)
	s.SetConfig(config.ServiceConfig{
		Name:    name,
		Type:    config.TypeSimple,
		Restart: config.RestartNever,
		ExecStart: []config.Exec{
			{
				Cmd:  "foo",
				Args: []string{"--bar"},
			},
		},
		StartLimitBurst: 5,
		StartLimitDur:   1 * time.Millisecond,
	})
	s.logger = log.Default()

	return s
}

type testCommand struct {
	cmd               string
	args              []string
	status            chan goCmd.Status
	commandStartFunc  func(c string, a []string) error
	commandStopStatus goCmd.Status
}

func (c *testCommand) Status() goCmd.Status {
	return goCmd.Status{}
}

func (c *testCommand) Start(env []string) (err error) {
	c.status = make(chan goCmd.Status)
	return c.commandStartFunc(c.cmd, c.args)
}

func (c *testCommand) StartNotify(env []string, sockPath string, startTimeoutDuration time.Duration) (err error) {
	return nil
}

func (c *testCommand) Stop(timeout time.Duration) error {
	go func() {
		c.status <- c.commandStopStatus
	}()
	return nil
}

func (c *testCommand) Streams() *cmd.Streams {
	return &cmd.Streams{
		Status: c.status,
	}
}

func helperWaitForState(getStatus chan chan api.ServiceStatus, state api.ServiceState) (err error) {

	start := time.Now()

	response := make(chan api.ServiceStatus)

	timeout := 2 * time.Second

loop:
	for {

		if time.Since(start) > timeout {
			return fmt.Errorf("timed out waiting for expected status")
		}

		select {
		case getStatus <- response:
			select {
			case d := <-response:
				if d.State == state {
					break loop
				}
			case <-time.After(timeout - time.Since(start)):
				return fmt.Errorf("timed out waiting for expected status")
			}
		case <-time.After(timeout - time.Since(start)):
			return fmt.Errorf("timed out waiting for expected status")
		}
	}
	return
}

func TestMain(m *testing.M) {
	cleanup, err := superdTesting.SetupEnv()
	if err != nil {
		panic(err)
	}

	ret := m.Run()

	if err := cleanup(); err != nil {
		panic(err)
	}

	os.Exit(ret)
}

func TestServiceState(t *testing.T) {
	subtests := []struct {
		name      string
		inName    string
		expected  api.ServiceState
		startFunc func(c string, a []string) error
	}{
		{
			name:     "state starting",
			inName:   "testStatusStarting",
			expected: api.ServiceStarting,
			startFunc: func(c string, a []string) error {
				// sleep for some arbitrary time so that the "starting" state
				// has enough time to be received. if this is too short, it may
				// transistion to "started" before the state we want is sent
				time.Sleep(3 * time.Second)
				return nil
			},
		},
		{
			name:      "state started",
			inName:    "testStatusStarted",
			expected:  api.ServiceStarted,
			startFunc: func(c string, a []string) error { return nil },
		},
		{
			name:      "state failed",
			inName:    "testStatusFailed",
			expected:  api.ServiceFailed,
			startFunc: func(c string, a []string) error { return cmd.ErrCmdExitedNonZero{Ret: -1} },
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {
			quit := make(chan struct{})

			s := helperService(subtest.inName)
			startFunc := subtest.startFunc

			s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
				return &testCommand{
					cmd:              exe,
					args:             args,
					commandStartFunc: startFunc,
				}
			}
			c := s.Config()
			// the retry mechanism makes test intermittently time out
			// (depending on system load, etc), so disable it
			c.StartLimitBurst = 1
			s.SetConfig(c)

			go s.Run(quit)

			select {
			case s.Start <- true:
			case <-time.After(1 * time.Second):
				t.Fatalf("timed out trying to start test process")
			}

			if err := helperWaitForState(s.getStatus, subtest.expected); err != nil {
				t.Fatal(err)
			}

			close(quit)
		})
	}
}

// testing the "stopped" state requires a bit more setup
func TestServiceStateStopped(t *testing.T) {

	quit := make(chan struct{})

	s := helperService("testServiceStateStopped")
	s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
		return &testCommand{
			cmd:              exe,
			args:             args,
			commandStartFunc: func(cmd string, args []string) error { return nil },
			commandStopStatus: goCmd.Status{
				Complete: true,
				Exit:     0,
				Error:    nil,
				StopTs:   999999,
			},
		}
	}
	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	// wait for "started" state
	if err := helperWaitForState(s.getStatus, api.ServiceStarted); err != nil {
		t.Fatal(err)
	}

	// signal stop
	select {
	case s.Stop <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to stop test process")
	}

	// wait for "stopped" state
	if err := helperWaitForState(s.getStatus, api.ServiceStopped); err != nil {
		t.Fatal(err)
	}

	close(quit)
}

// tests stopping a service with a non-zero ret code
func TestServiceStateStoppedFail(t *testing.T) {

	quit := make(chan struct{})

	s := helperService("testServiceStateStoppedFail")
	s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
		return &testCommand{
			cmd:              exe,
			args:             args,
			commandStartFunc: func(cmd string, args []string) error { return nil },
			commandStopStatus: goCmd.Status{
				Complete: false,
				Exit:     -1,
				Error:    nil,
				StopTs:   999999,
			},
		}
	}
	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	// wait for "started" state
	if err := helperWaitForState(s.getStatus, api.ServiceStarted); err != nil {
		t.Fatal(err)
	}

	// signal stop
	select {
	case s.Stop <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to stop test process")
	}

	// wait for "stopped" state, *not* "failed"
	if err := helperWaitForState(s.getStatus, api.ServiceStopped); err != nil {
		t.Fatal(err)
	}

	close(quit)
}

// This checks that the service restart mechanism is not triggered if the
// startup retry loop fails. It also checks that the execOnFailure is still
// executed.
func TestNoRestartOnStartFailure(t *testing.T) {

	quit := make(chan struct{})

	execStartCmd := "startCmd"
	execFailCmd := "failCmd"

	expectedTries := 4
	var tries int32
	atomic.StoreInt32(&tries, 0)

	getRestartTimer = func(_ time.Duration) <-chan time.Time {
		t.Fatalf("unexpected restart requested")
		return time.After(1 * time.Millisecond)
	}

	s := helperService("testStartupRetry")
	// newCommand here returns the ExecStart cmd and simulates a failure so
	// that ExecOnFailure is triggered.
	s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
		return &testCommand{
			cmd:  exe,
			args: args,
			commandStartFunc: func(c string, _ []string) error {
				switch c {
				case execStartCmd:
					atomic.AddInt32(&tries, 1)
					// fake the service failure
					return cmd.ErrCmdExitedNonZero{Ret: -1}
				case execFailCmd:
				default:
					t.Fatalf("unexpected command: %s\n", c)
				}
				return nil
			},
		}
	}

	c := s.Config()
	c.ExecOnFailure = []config.Exec{
		{
			Cmd: execFailCmd,
		},
	}
	c.ExecStart = []config.Exec{
		{
			Cmd:  execStartCmd,
			Args: []string{},
		},
	}
	c.StartLimitBurst = expectedTries
	// Limit duration should be high enough so that the duration doesn't expire
	// within testing
	c.StartLimitDur = 10 * time.Second
	c.Restart = config.RestartAlways
	s.SetConfig(c)

	done := make(chan struct{})
	go func() {
		s.Run(quit)
		close(done)
	}()

	select {
	case s.Start <- true:
	case <-time.After(2 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	if err := helperWaitForState(s.getStatus, api.ServiceFailed); err != nil {
		t.Fatal(err)
	}

	close(quit)

	tries = int32(atomic.LoadInt32(&tries))
	if int(atomic.LoadInt32(&tries)) != expectedTries {
		t.Fatalf("unexpected number of tries to start service: %d, expected: %d\n", tries, expectedTries)
	}

	select {
	case <-time.After(2 * time.Second):
		t.Fatalf("timed out waiting for service.Run to quit")
	case <-done:
	}
}

func TestExecOnFailure(t *testing.T) {

	quit := make(chan struct{})

	startCmd := "startCmd"
	startArgs := []string{"--lime"}
	expectedCmd := "foo"
	expectedArgs := []string{"-t", "bar", "--bazz"}
	expectedRun := make(chan struct{})

	s := helperService("testExecOnFailure")
	// newCommand here returns the ExecStart cmd and simulates a failure so
	// that ExecOnFailure is triggered
	s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
		return &testCommand{
			cmd:  exe,
			args: args,
			commandStartFunc: func(c string, a []string) error {
				if c == startCmd {
					// fake the service failure
					return cmd.ErrCmdExitedNonZero{Ret: -1}
				} else if c == expectedCmd {
					if !reflect.DeepEqual(a, expectedArgs) {
						t.Fatalf("expected: %q, got: %q\n", expectedArgs, a)
					} else {
						close(expectedRun)
					}
					return cmd.ErrCmdExited{}
				}
				t.Fatalf("expected: %s, got: %s\n", expectedCmd, c)
				return nil
			},
		}
	}

	c := s.Config()
	c.ExecOnFailure = []config.Exec{
		{
			Cmd:  expectedCmd,
			Args: expectedArgs,
		},
	}
	c.ExecStart = []config.Exec{
		{
			Cmd:  startCmd,
			Args: startArgs,
		},
	}
	// the retry mechanism makes testing this more difficult, so disable it
	c.StartLimitBurst = 1
	s.SetConfig(c)

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	select {
	case <-expectedRun:
	case <-time.After(2 * time.Second):
		t.Fatal("timed out waiting for ExecOnFailure to run")
	}

	close(quit)
}

func TestExecStartPre(t *testing.T) {
	type exes struct {
		list []config.Exec
		*sync.Mutex
	}

	// each ExecStartPre cmd + args is appended to the list here as they are
	// run, for comparison/validation at the end
	type runCmds struct {
		list []string
		*sync.Mutex
	}

	subtests := []struct {
		name string
		in   exes
	}{
		{
			name: "only one",
			in: exes{
				list: []config.Exec{
					{
						Cmd:  "do_stuff_pre",
						Args: []string{"-a", "thing", "--mango"},
					},
				},
				Mutex: new(sync.Mutex),
			},
		},
		{
			name: "multiple execs",
			in: exes{
				list: []config.Exec{
					{
						Cmd:  "do_stuff_pre",
						Args: []string{"-a", "thing", "--mango"},
					},
					{
						Cmd:  "mkdir",
						Args: []string{"/tmp/foo", "-p"},
					},
					{
						Cmd:  "echo",
						Args: []string{"-n", "\"look at me\""},
					},
				},
				Mutex: new(sync.Mutex),
			},
		},
	}

	for _, subtest := range subtests {
		t.Run(subtest.name, func(t *testing.T) {

			subtestDone := make(chan struct{})
			out := runCmds{[]string{}, new(sync.Mutex)}
			quit := make(chan struct{})

			s := helperService("testExecStartPre")
			s.newCommand = func(e string, args []string, outFile *os.File) cmd.Commander {
				return &testCommand{
					cmd:  e,
					args: args,
					commandStartFunc: func(c string, a []string) error {
						out.Lock()
						defer out.Unlock()
						subtest.in.Lock()
						defer subtest.in.Unlock()

						expectedSize := len(subtest.in.list)
						size := len(out.list)
						switch {
						case size == expectedSize:
							close(subtestDone)
						case size > expectedSize:
							// probably running ExecStart
						default:
							out.list = append(out.list, fmt.Sprintf("%s %s", c, a))
						}
						return nil
					},
				}
			}

			c := s.Config()
			c.ExecStartPre = subtest.in.list
			s.SetConfig(c)

			go s.Run(quit)

			select {
			case s.Start <- true:
			case <-time.After(1 * time.Second):
				t.Fatalf("timed out trying to start test process")
			}

			select {
			case <-subtestDone:
			case <-time.After(2 * time.Second):
				t.Fatal("timed out waiting for ExecStartPre to run")
			}

			var expected []string
			for _, e := range subtest.in.list {
				expected = append(expected, fmt.Sprintf("%s %s", e.Cmd, e.Args))
			}

			subtest.in.Lock()
			out.Lock()
			if !reflect.DeepEqual(out.list, expected) {
				t.Fatalf("Expected: %s, got: %s", expected, out.list)
			}

			close(quit)
		})
	}
}

func TestFailureRestart(t *testing.T) {

	quit := make(chan struct{})
	expectedRun := make(chan struct{})
	var totalRuns int32
	atomic.StoreInt32(&totalRuns, 0)

	s := helperService("testFailureRestart")
	s.newCommand = func(exe string, args []string, out *os.File) cmd.Commander {
		return &testCommand{
			cmd:  exe,
			args: args,
			commandStartFunc: func(c string, a []string) error {
				// fake the service failure
				atomic.AddInt32(&totalRuns, 1)
				if int(atomic.LoadInt32(&totalRuns)) == s.Config().StartLimitBurst {
					close(expectedRun)
				}
				return cmd.ErrCmdExitedNonZero{Ret: -1}
			},
		}
	}
	c := s.Config()
	c.Restart = config.RestartFailure
	s.SetConfig(c)

	go s.Run(quit)

	select {
	case s.Start <- true:
	case <-time.After(1 * time.Second):
		t.Fatalf("timed out trying to start test process")
	}

	select {
	case <-expectedRun:
	case <-time.After(2 * time.Second):
		t.Fatal("timed out waiting for ExecStartPre to run")
	}

	close(quit)
}

func TestShouldRestart(t *testing.T) {
	subtests := []struct {
		name      string
		inStatus  api.ServiceState
		inType    config.Type
		inRestart config.RestartCond
		expected  bool
	}{
		{
			name:      "simple failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "simple failed never restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartNever,
			expected:  false,
		},
		{
			name:      "simple failed restart on failure",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeSimple,
			inRestart: config.RestartFailure,
			expected:  true,
		},
		{
			name:      "simple starting always restart",
			inStatus:  api.ServiceStarting,
			inType:    config.TypeSimple,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot stopped always restart",
			inStatus:  api.ServiceStopped,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot failed always restart",
			inStatus:  api.ServiceFailed,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  true,
		},
		{
			name:      "oneshot started always restart",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartAlways,
			expected:  false,
		},
		{
			name:      "oneshot started restart on failure",
			inStatus:  api.ServiceStarted,
			inType:    config.TypeOneshot,
			inRestart: config.RestartFailure,
			expected:  false,
		},
	}

	for _, st := range subtests {
		t.Run(st.name, func(t *testing.T) {
			s := helperService(st.name)
			c := s.Config()
			c.Type = st.inType
			c.Restart = st.inRestart
			s.SetConfig(c)

			var out bool
			switch st.inType {
			case config.TypeSimple:
				out = s.simpleShouldRestart(st.inStatus)
			case config.TypeOneshot:
				out = s.oneshotShouldRestart(st.inStatus)
			default:
				t.Fatalf("unexpected type %s", st.inType)
			}
			if out != st.expected {
				t.Fatalf("expected: %t, received: %t", st.expected, out)
			}
		})
	}
}

func TestBeforeAfter(t *testing.T) {
	subtests := []struct {
		name string
		in   []string
	}{
		{
			name: "before / after",
			in:   []string{"foo", "banana", "something"},
		},
	}

	for _, st := range subtests {

		t.Run(st.name, func(t *testing.T) {
			s := helperService("test")
			for _, i := range st.in {
				s.AddAfter(i)
				s.AddBefore(i)
			}

			out := s.After()
			if !reflect.DeepEqual(out, st.in) {
				t.Fatalf("Expected: %q, got: %q", st.in, out)
			}

			out = s.Before()
			if !reflect.DeepEqual(out, st.in) {
				t.Fatalf("Expected: %q, got: %q", st.in, out)
			}
		})
	}
}
