// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package testing

import (
	"os"
	"path/filepath"
)

// SetupEnv creates a temp dir structure for and sets env. variables to point
// to the temp dirs. It returns a function that should be called to clean up
// after testing is complete.
func SetupEnv() (RestoreEnv func() error, err error) {
	oldEnv := make(map[string]string)
	vars := []string{"XDG_RUNTIME_DIR", "XDG_STATE_HOME", "XDG_DATA_DIRS", "XDG_CONFIG_HOME"}
	var tempDir string

	tempDir, err = os.MkdirTemp("", "superd")
	if err != nil {
		return
	}

	for _, name := range vars {
		// backup old env
		v, ok := os.LookupEnv(name)
		if ok {
			oldEnv[name] = v
		}

		path := filepath.Join(tempDir, name)
		if err = os.MkdirAll(path, 0750); err != nil {
			return
		}

		if err = os.Setenv(name, path); err != nil {
			return
		}
	}

	// PATH is special, since it's prepended to the existing env var and not
	// overwriting it
	existingPATH, ok := os.LookupEnv("PATH")
	if ok {
		oldEnv["PATH"] = existingPATH
	}

	var scriptdir string
	if scriptdir, err = filepath.Abs("../../test"); err != nil {
		return
	}

	if err = os.Setenv("PATH", scriptdir+":"+existingPATH); err != nil {
		return
	}

	RestoreEnv = func() (err error) {
		err = os.RemoveAll(tempDir)
		if err != nil {
			return
		}

		// restore old env
		for name, val := range oldEnv {
			if err = os.Setenv(name, val); err != nil {
				return
			}
		}
		return
	}

	return
}
