// Copyright 2022 Clayton Craft <clayton@craftyguy.net>
// SPDX-License-Identifier: GPL-3.0-or-later

package api

import "encoding/json"

type ServiceState string

const (
	ServiceFailed   ServiceState = "failed"
	ServiceStarted  ServiceState = "started"
	ServiceStarting ServiceState = "starting"
	ServiceStopped  ServiceState = "stopped"
)

type ServiceStatus struct {
	State ServiceState
	Pid   int

	// path to the service's config file
	ConfigPath string

	// whether or not the service is autostarted
	Autostart bool

	// service "type", e.g. "oneshot"
	Type string

	// number of seconds that the service has been started, not useful if
	// Status is *not* started
	Uptime int64

	// short description of the service
	Description string
}

type ReqCommand string

const (
	CmdDisable ReqCommand = "disable"
	CmdEnable  ReqCommand = "enable"
	CmdReload  ReqCommand = "reload"
	CmdRestart ReqCommand = "restart"
	CmdSetEnv  ReqCommand = "set-env"
	CmdStart   ReqCommand = "start"
	CmdStatus  ReqCommand = "status"
	CmdStop    ReqCommand = "stop"
)

type Request struct {
	Cmd    ReqCommand `json:"cmd"`
	Params []string   `json:"params"`
}

func (r Request) Json() (req []byte, err error) {
	return json.Marshal(r)
}

type Response interface {
	Json() ([]byte, error)
}

type ErrorResponse struct {
	Error    bool
	ErrorMsg string
}

func NewErrorResponse(message string) ErrorResponse {
	return ErrorResponse{
		Error:    true,
		ErrorMsg: message,
	}
}

type NilResponse struct {
	ErrorResponse `json:"error"`
}

func (r NilResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}

type ServiceStatusResponse struct {
	Result        map[string]ServiceStatus `json:"result"`
	ErrorResponse `json:"error"`
}

func (r ServiceStatusResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}

type StringResponse struct {
	Result        string `json:"result"`
	ErrorResponse `json:"error"`
}

func (r StringResponse) Json() (resp []byte, err error) {
	return json.Marshal(r)
}
